# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
import subprocess
from contextlib import contextmanager
from dataclasses import dataclass
from copy import deepcopy
from pathlib import Path
from typing import Any, Tuple, Optional, Union
import shutil
import tempfile
import time
from loguru import logger
import yaml
from msserviceprofiler.modelevalstate.config.config import MindieConfig, VllmConfig, OptimizerConfigField, KubectlConfig
from msserviceprofiler.modelevalstate.config.base_config import simulate_flag, SIMULATE
from msserviceprofiler.modelevalstate.config.custom_command import MindieCommand, VllmCommand
from msserviceprofiler.modelevalstate.optimizer.custom_process import CustomProcess
from msserviceprofiler.modelevalstate.optimizer.utils import backup, remove_file, close_file_fp
from msserviceprofiler.msguard.security import open_s


@dataclass
class ConfigContextdict:
    origin_config: dict
    cur_key: str
    next_key: str
    next_level: str
    value: Any
    current_depth: int


@dataclass
class ConfigContextlist:
    origin_config: list
    cur_key: str
    next_key: str
    next_level: str
    value: Any
    current_depth: int


class Simulator(CustomProcess):
    def __init__(self, mindie_config: MindieConfig, bak_path: Optional[Path] = None,
                 print_log: bool = False):
        super().__init__(bak_path=bak_path, print_log=print_log, process_name=mindie_config.process_name)
        self.mindie_config = mindie_config
        logger.debug(f"config path {self.mindie_config.config_path!r}", )
        if not self.mindie_config.config_path.exists():
            raise FileNotFoundError(self.mindie_config.config_path)
        with open_s(self.mindie_config.config_path, "r") as f:
            data = json.load(f)
        self.default_config = data
        logger.debug(f"config bak path {self.mindie_config.config_bak_path!r}", )
        if self.mindie_config.config_bak_path.exists():
            self.mindie_config.config_bak_path.unlink()
        with open_s(self.mindie_config.config_bak_path, 'w') as fout:
            json.dump(self.default_config, fout, indent=4)
        self.command = MindieCommand(self.mindie_config.command).command

    @staticmethod
    def is_int(x):
        try:
            int(x)
            return True
        except ValueError:
            return False
        
    @staticmethod
    def set_config_for_dict(context: ConfigContextdict):
        if context.cur_key in context.origin_config:
            Simulator.set_config(context.origin_config[context.cur_key], context.next_level, context.value, 
                                 context.current_depth)
        elif Simulator.is_int(context.cur_key):
            raise KeyError(f"data: {context.origin_config}, key: {context.cur_key}")
        elif Simulator.is_int(context.next_key):
            context.origin_config[context.cur_key] = []
            Simulator.set_config(context.origin_config[context.cur_key], context.next_level, context.value, 
                                 context.current_depth)
        else:
            context.origin_config[context.cur_key] = {}
            Simulator.set_config(context.origin_config[context.cur_key], context.next_level, context.value, 
                                 context.current_depth)

    @staticmethod
    def set_config_for_list(context: ConfigContextlist):
        if len(context.origin_config) > int(context.cur_key):
            Simulator.set_config(context.origin_config[int(context.cur_key)], context.next_level, context.value, 
                                 context.current_depth)
        elif len(context.origin_config) == int(context.cur_key) and Simulator.is_int(context.next_key):
            context.origin_config.append([])
            Simulator.set_config(context.origin_config[int(context.cur_key)], context.next_level, context.value, 
                                 context.current_depth)
        elif len(context.origin_config) == int(context.cur_key) and not Simulator.is_int(context.next_key):
            context.origin_config.append({})
            Simulator.set_config(context.origin_config[int(context.cur_key)], context.next_level, context.value, 
                                 context.current_depth)
        else:
            raise IndexError(f"data: {context.origin_config}, index: {context.cur_key}")

    @staticmethod
    def set_config(origin_config, key: str, value: Any, current_depth=0):
        if current_depth > 10:
            raise RecursionError("Exceeded maximum recursion depth")
        next_level = None
        try:
            if "." in key:
                _f_index = key.index(".")
                _cur_key, next_level = key[:_f_index], key[_f_index + 1:]
            else:
                _cur_key = key
            if next_level is None:
                if isinstance(origin_config, dict):
                    origin_config[_cur_key] = value
                elif isinstance(origin_config, list):
                    if len(origin_config) > int(_cur_key):
                        origin_config[int(_cur_key)] = value
                    else:
                        origin_config.append(value)
                return
            if "." in next_level:
                _next_index = next_level.index(".")
                _next_key = next_level[:_next_index]
            elif next_level:
                _next_key = next_level
            else:
                _next_key = None
        except Exception as e:
            logger.error(f"Unexpected error occurred at {key}")
            raise e
        if isinstance(origin_config, dict):
            context = ConfigContextdict(
                origin_config=origin_config,
                cur_key=_cur_key,
                next_key=_next_key,
                next_level=next_level,
                value=value,
                current_depth=current_depth + 1
            )
            Simulator.set_config_for_dict(context)
        elif isinstance(origin_config, list):
            context = ConfigContextlist(
                origin_config=origin_config,
                cur_key=_cur_key,
                next_key=_next_key,
                next_level=next_level,
                value=value,
                current_depth=current_depth + 1
            )
            Simulator.set_config_for_list(context)
        else:
            raise ValueError(f"Not Support type {type(origin_config)}")

    def update_command(self):
        self.command = MindieCommand(self.mindie_config.command).command
    
    def before_run(self, run_params: Optional[Tuple[OptimizerConfigField]] = None):
        # 根据params 修改配置文件
        # 启动mindie仿真
        self.update_command()
        self.update_config(run_params)
        super().before_run(run_params)
        subprocess.run(["pkill", "-9", "mindie"], env=self.env, stdout=self.run_log_fp,
                       stderr=subprocess.STDOUT, text=True, cwd=self.work_path)
        subprocess.run(["npu-smi", "info"], env=self.env, stdout=self.run_log_fp,
                       stderr=subprocess.STDOUT, text=True, cwd=self.work_path)

    def backup(self):
        super().backup()
        backup(self.mindie_config.config_path, self.bak_path, self.__class__.__name__)

    def update_config(self, params: Optional[Tuple[OptimizerConfigField]] = None):
        # 将params值更新到新的config中
        if not params:
            return
        new_config = deepcopy(self.default_config)
        for p in params:
            if not p.config_position.startswith("BackendConfig"):
                continue
            Simulator.set_config(new_config, p.config_position, p.value)

        # 将新的config写入到config文件中
        logger.debug(f"new config {new_config}")
        if self.mindie_config.config_path.exists():
            self.mindie_config.config_path.unlink()
        with open_s(self.mindie_config.config_path, "w") as fout:
            json.dump(new_config, fout, indent=4, ensure_ascii=False)

    def check_success(self, print_log=False):
        output = self.get_log()
        if self.print_log:
            logger.info(output)
        if output and "Daemon start success!" in output:
            return True
        if self.process.poll() is None:
            return False
        elif self.process.poll() == 0:
            return True
        else:
            raise subprocess.SubprocessError(
                f"Failed in run {self.command}. return code: {self.process.returncode}. log: {self.run_log}")

    def stop(self, del_log: bool = True):
        # 恢复默认的mindie 配置
        remove_file(self.mindie_config.config_path)
        with open_s(self.mindie_config.config_path, "w") as fout:
            json.dump(self.default_config, fout, indent=4, ensure_ascii=False)
        super().stop(del_log)


class DisaggregationSimulator(CustomProcess):
    from msserviceprofiler.modelevalstate.config.custom_command import KubectlCommand

    def __init__(self, mindie_config: KubectlConfig, bak_path: Optional[Path] = None):
        super().__init__(bak_path=bak_path)
        self.mindie_config = mindie_config
        logger.info(f"config path {self.mindie_config.config_single_path}", )
        if not self.mindie_config.config_single_path.exists():
            raise FileNotFoundError(self.mindie_config.config_single_path)
        with open_s(self.mindie_config.config_single_path, "r") as f:
            data = json.load(f)
        with open_s(self.mindie_config.config_single_pd_path, "r") as f:
            pd_data = json.load(f)
        self.default_pd_config = pd_data
        self.default_config = data
        logger.info(f"config bak path {self.mindie_config.config_single_bak_path}", )
        if self.mindie_config.config_single_bak_path.exists():
            self.mindie_config.config_single_bak_path.unlink()
        with open_s(self.mindie_config.config_single_bak_path, "w") as fout:
            json.dump(self.default_config, fout, indent=4)
        self.run_log = None
        self.mindie_log_offset = 0
        self.bak_path = bak_path
        self.mindie_log_fp = None
        self.process = None
        self.command = self.KubectlCommand(self.mindie_config.command).command
        self.log_command = self.KubectlCommand(self.mindie_config.command).log_command
        self.monitor_command = self.KubectlCommand(self.mindie_config.command).monitor_command

    @staticmethod
    def is_int(x):
        try:
            int(x)
            return True
        except ValueError:
            return False

    @staticmethod
    def set_config_for_dict(context: ConfigContextdict):
        if context.cur_key in context.origin_config:
            DisaggregationSimulator.set_config(context.origin_config[context.cur_key], context.next_level, 
                                               context.value, context.current_depth)
        elif DisaggregationSimulator.is_int(context.cur_key):
            raise KeyError(f"data: {context.origin_config}, key: {context.cur_key}")
        elif DisaggregationSimulator.is_int(context.next_key):
            context.origin_config[context.cur_key] = []
            DisaggregationSimulator.set_config(context.origin_config[context.cur_key], context.next_level, 
                                               context.value, context.current_depth)
        else:
            context.origin_config[context.cur_key] = {}
            DisaggregationSimulator.set_config(context.origin_config[context.cur_key], context.next_level, 
                                               context.value, context.current_depth)

    @staticmethod
    def set_config_for_list(context: ConfigContextlist):
        if len(context.origin_config) > int(context.cur_key):
            DisaggregationSimulator.set_config(context.origin_config[int(context.cur_key)], context.next_level, 
                                               context.value, context.current_depth)
        elif len(context.origin_config) == int(context.cur_key) and DisaggregationSimulator.is_int(context.next_key):
            context.origin_config.append([])
            DisaggregationSimulator.set_config(context.origin_config[int(context.cur_key)], context.next_level, 
                                               context.value, context.current_depth)
        elif len(context.origin_config) == int(context.cur_key) and not \
            DisaggregationSimulator.is_int(context.next_key):
            context.origin_config.append({})
            DisaggregationSimulator.set_config(context.origin_config[int(context.cur_key)], context.next_level, 
                                               context.value, context.current_depth)
        else:
            raise IndexError(f"data: {context.origin_config}, index: {context.cur_key}")

    @staticmethod
    def set_config(origin_config, key: str, value: Any, current_depth=0):
        if current_depth > 10:
            raise RecursionError("Exceeded maximum recursion depth")
        next_level = None
        try:
            if "." in key:
                _f_index = key.index(".")
                _cur_key, next_level = key[:_f_index], key[_f_index + 1:]
            else:
                _cur_key = key
            if next_level is None:
                if isinstance(origin_config, dict):
                    origin_config[_cur_key] = value
                elif isinstance(origin_config, list):
                    if len(origin_config) > int(_cur_key):
                        origin_config[int(_cur_key)] = value
                    else:
                        origin_config.append(value)
                return
            if "." in next_level:
                _next_index = next_level.index(".")
                _next_key = next_level[:_next_index]
            elif next_level:
                _next_key = next_level
            else:
                _next_key = None
        except Exception as e:
            logger.error(f"Unexpected error occurred at {key}")
            raise e
        if isinstance(origin_config, dict):
            context = ConfigContextdict(
                origin_config=origin_config,
                cur_key=_cur_key,
                next_key=_next_key,
                next_level=next_level,
                value=value,
                current_depth=current_depth + 1
            )
            DisaggregationSimulator.set_config_for_dict(context)
        elif isinstance(origin_config, list):
            context = ConfigContextlist(
                origin_config=origin_config,
                cur_key=_cur_key,
                next_key=_next_key,
                next_level=next_level,
                value=value,
                current_depth=current_depth + 1
            )
            DisaggregationSimulator.set_config_for_list(context)
        else:
            raise ValueError(f"Not Support type {type(origin_config)}")
    
    def prepare_before_start_server(self):
        bash_path = shutil.which("bash")
        if bash_path is not None:
            subprocess.run([bash_path, self.mindie_config.delete_path, "mindie", "."], 
                           cwd=self.mindie_config.kubectl_default_path)
            while True:
                singal = True
                proc = subprocess.run(self.log_command, stdout=subprocess.PIPE, text=True, 
                                      cwd=self.mindie_config.kubectl_default_path)
                lines = proc.stdout.splitlines()
                for line in lines:
                    if line.startswith('mindie'):
                        singal = False
                if singal is True:
                    break
                time.sleep(1)
        else:
            logger.error("bash not found in path")

    def backup(self, del_log=True):
        backup(self.mindie_config.config_single_path, self.bak_path, self.__class__.__name__)
        if not del_log and self.run_log:
            backup(self.run_log, self.bak_path, self.__class__.__name__)

    def update_config(self, params: Tuple[OptimizerConfigField]):
        # 将params值更新到新的config中
        new_config = deepcopy(self.default_config)
        pd_config = deepcopy(self.default_pd_config)
        for p in params:
            if p.config_position.startswith("default"):
                DisaggregationSimulator.set_config(pd_config, p.config_position, p.value)
            if not p.config_position.startswith("BackendConfig"):
                continue
            DisaggregationSimulator.set_config(new_config, p.config_position, p.value)

        # 将新的config写入到config文件中
        logger.debug(f"new config {new_config}")
        if self.mindie_config.config_single_path.exists():
            self.mindie_config.config_single_path.unlink()
        with open_s(self.mindie_config.config_single_path, "w") as fout:
            json.dump(new_config, fout, indent=4)
        if self.mindie_config.config_single_pd_path.exists():
            self.mindie_config.config_single_pd_path.unlink()
        with open_s(self.mindie_config.config_single_pd_path, "w") as fout:
            json.dump(pd_config, fout, indent=4)

    def test_curl(self):
        import requests
        curl_port = None
        yaml_dir = self.mindie_config.kubectl_single_path.parent
        yaml_path = os.path.join(yaml_dir, "deployment/mindie_service_single_container.yaml")
        with open_s(yaml_path, 'r') as file:
            all_documents = yaml.safe_load_all(file)
            for doc in all_documents:
                # 检查文档中是否存在 spec.ports 部分
                if 'spec' in doc and 'ports' in doc['spec']:
                    ports = doc['spec']['ports']
                    for port in ports:
                        if 'nodePort' in port:
                            curl_port = port['nodePort']
        if curl_port:
            url = f"http://127.0.0.1:{curl_port}"
        else:
            raise ("cannot find port from mindie_service_single_container.yaml, please check")

        # 定义请求体的 JSON 数据
        payload = {
            "inputs": "Please introduce yourself.",
            "parameters": {
                "max_new_tokens": 20,
                "temperature": 0.3,
                "top_p": 0.3,
                "top_k": 5,
                "do_sample": True,
                "repetition_penalty": 1.05,
                "seed": 128
            }
        }

        # 定义请求头（如果需要）
        headers = {
            "Content-Type": "application/json"
        }

        try:
            # 发送 POST 请求
            response = requests.post(url, json=payload, headers=headers, timeout=10)
            
            # 检查响应状态码
            if response.status_code == 200:
                return True
            else:
                return False
        except requests.exceptions.RequestException as e:
            return False

    def check_success(self, print_log=False):
        with open_s(self.run_log, "r") as f:
            try:
                f.seek(self.mindie_log_offset)
                output = f.read()
                self.mindie_log_offset = f.tell()
            except Exception as e:
                logger.info(f"Failed in read mindie log. error: {e}")
        if output:
            if print_log:
                logger.info(f"simulate out: \n{output}")
            if "MindIE-MS coordinator is ready!!!" in output:
                while True:
                    if self.test_curl() is True:
                        return True
                    time.sleep(1)
        return False

    def start_server(self, run_params: Tuple[OptimizerConfigField]):
        self.prepare_before_start_server()
        self.mindie_log_fp, self.run_log = tempfile.mkstemp(prefix="modelevalstate_mindie")
        self.mindie_log_offset = 0
        if self.mindie_config.work_path:
            cwd = self.mindie_config.work_path
        else:
            cwd = os.getcwd()
        logger.info(f"self.command: {self.command}")
        self.process = subprocess.run(self.command, env=self.env, text=True, 
                                      cwd=self.mindie_config.kubectl_default_path)
        logger.info(f"self.log_command: {self.log_command}")
        while True:
            proc = subprocess.run(self.log_command, stdout=subprocess.PIPE, text=True, 
                                cwd=self.mindie_config.kubectl_default_path)
            
            lines = proc.stdout.splitlines()
            for line in lines:
                if line.startswith('mindie'):
                    mindie_name = line.split()
                    break
            if mindie_name[3] == 'Running':
                break
            time.sleep(1)
        kubectl_monitor_command = self.KubectlCommand(self.mindie_config.command).monitor_command
        kubectl_monitor_command.append(mindie_name[1])
        logger.info(f"mindie_name: {mindie_name[1]}")
        self.log_process = subprocess.Popen(kubectl_monitor_command, stdout=self.mindie_log_fp, 
                                            stderr=subprocess.STDOUT, env=self.env, text=True, 
                                            cwd=self.mindie_config.kubectl_default_path)
        logger.info(f"command: {' '.join(kubectl_monitor_command)}, log file: {self.run_log}")

    def run(self, run_params: Tuple[OptimizerConfigField]):
        logger.info(f'start run in simulator. run params: {run_params}')
        # 根据params 修改配置文件
        self.update_config(run_params)
        # 启动mindie仿真
        self.start_server(run_params)

    def stop(self, del_log=True):
        logger.info("Stop simulator process")
        if self.bak_path:
            self.backup()
        close_file_fp(self.mindie_log_fp)
        if del_log:
            remove_file(self.run_log)
        self.mindie_log_offset = 0
        try:
            bash_path = shutil.which("bash")
            if bash_path is not None:
                subprocess.run([bash_path, self.mindie_config.delete_path, "mindie", "./"])
            else:
                logger.error("bash not found in path")
        except Exception as e:
            logger.error(f"Failed to stop simulator process. {e}")


class VllmSimulator(CustomProcess):
    def __init__(self, vllm_config: VllmConfig, bak_path: Optional[Path] = None,
                 print_log: bool = False):
        super().__init__(bak_path=bak_path, print_log=print_log, process_name=vllm_config.process_name)
        self.vllm_config = vllm_config
        self.command = VllmCommand(vllm_config.command).command

    def update_command(self):
        self.command = VllmCommand(self.vllm_config.command).command

    def backup(self):
        super().backup()
        backup(self.vllm_config.output, self.bak_path, self.__class__.__name__)

    def before_run(self, run_params: Optional[Tuple[OptimizerConfigField]] = None):
        self.update_command()
        super().before_run(run_params)
        subprocess.run(["pkill", "-15", "vllm"], env=self.env, stdout=self.run_log_fp,
                       stderr=subprocess.STDOUT, text=True, cwd=self.work_path)
        subprocess.run(["npu-smi", "info"], env=self.env, stdout=self.run_log_fp,
                       stderr=subprocess.STDOUT, text=True, cwd=self.work_path)

    def check_success(self, print_log=False):
        output = self.get_log()
        if self.print_log:
            logger.info(output)
        if output and "Application startup complete." in output:
            return True
        if self.process.poll() is None:
            return False
        elif self.process.poll() == 0:
            return True
        else:
            raise subprocess.SubprocessError(
                f"Failed in run {self.command}. return code: {self.process.returncode}. log: {self.run_log}")


@contextmanager
def enable_simulate(simulate):
    if simulate_flag and isinstance(simulate, Simulator):
        origin_data = simulate.default_config
        data = deepcopy(origin_data)
        simulate.default_config = data
        model_config = data["BackendConfig"]["ModelDeployConfig"]["ModelConfig"][0]
        if "plugin_params" in model_config:
            _plugin_params = json.loads(model_config["plugin_params"])
            if SIMULATE not in _plugin_params["plugin_type"]:
                _plugin_params["plugin_type"] += "," + SIMULATE
                model_config["plugin_params"] = json.dumps(_plugin_params)
        else:
            model_config["plugin_params"] = json.dumps({"plugin_type": SIMULATE})
        with open_s(simulate.mindie_config.config_path, 'w') as f:
            json.dump(data, f, indent=4)
        yield simulate_flag
        if simulate.mindie_config.config_path.exists():
            simulate.mindie_config.config_path.unlink()
        with open_s(simulate.mindie_config.config_path, 'w') as f:
            json.dump(origin_data, f, indent=4)
    else:
        yield simulate_flag
    return


# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import time

from ms_service_profiler import Profiler, Level
from ..module_hook import vllm_hook


def prof_add_request(request_id, prompt, *args, **kwargs):
    # 记录请求进入系统的时间
    Profiler(Level.INFO).domain("Request").res(request_id).event("httpReq")
    Profiler(Level.INFO).domain("Request").res(request_id).event("tokenize")


# generate -> add_request -> schedule -> execute_model
# 在请求进入引擎时记录时间戳，用于后续计算队列等待时间
@vllm_hook(
    hook_points=[
        ("vllm.engine.llm_engine", "LLMEngine.add_request"),
        ("vllm.engine.async_llm_engine", "AsyncLLMEngine.add_request"),
    ],
    min_version="0.6.3",
    max_version="0.6.3",
)
def add_request_063(original_func, this, request_id, prompt, *args, **kwargs):
    prof_add_request(request_id, prompt, *args, **kwargs)
    return original_func(this, request_id, prompt, *args, **kwargs)


@vllm_hook(("vllm.engine.llm_engine", "LLMEngine.add_request"), min_version="0.8.4")
def add_request_084(original_func, this, request_id, prompt, *args, **kwargs):
    prof_add_request(request_id, prompt, *args, **kwargs)
    return original_func(this, request_id, prompt, *args, **kwargs)


@vllm_hook(("vllm.engine.async_llm_engine", "AsyncLLMEngine.add_request"), min_version="0.8.4")
async def add_request_async(original_func, this, request_id, prompt, *args, **kwargs):
    prof_add_request(request_id, prompt, *args, **kwargs)
    return original_func(this, request_id, prompt, *args, **kwargs)


@vllm_hook(("vllm.engine.llm_engine", "LLMEngine._process_model_outputs"), min_version="0.8.4")
def process_model_outputs(original_func, this, ctx, request_id=None, *args, **kwargs):
    if len(ctx.output_queue) == 0:
        return original_func(this, ctx, request_id, *args, **kwargs)
    outputs, metadata_list, scheduler_outputs, _, _, _, skip = ctx.output_queue[0]
    ret = original_func(this, ctx, request_id, *args, **kwargs)

    request_id_list = []
    for ii, (sch_seq_group, meta) in enumerate(zip(scheduler_outputs.scheduled_seq_groups, metadata_list)):
        if ii in skip:
            continue

        seq_group, seq_request_id = sch_seq_group.seq_group, meta.request_id
        request_id_list.append(seq_request_id)
        if seq_group.is_finished() and len(seq_group.seqs) > 0:
            cur_seq = seq_group.seqs[0]
            profiler_recv = Profiler(Level.INFO).domain("Request").res(seq_request_id)
            profiler_reply = Profiler(Level.INFO).domain("Request").res(seq_request_id)
            profiler_recv.metric("recvTokenSize", cur_seq.get_prompt_len()).event("httpRes")
            profiler_reply.metric("replyTokenSize", cur_seq.get_output_len()).event("httpRes")

    prof = Profiler(Level.INFO).domain("Request").res(request_id_list)
    prof.event("detokenize")
    return ret


@vllm_hook(("vllm.engine.llm_engine", "LLMEngine.validate_output"), min_version="0.6.3", max_version="0.6.3")
def validate_output(original_func, this, output, output_type):
    profiler_recv = Profiler(Level.INFO).domain("Request")
    profiler_reply = Profiler(Level.INFO).domain("Request")
    if output.finished is True:
        request_id = output.request_id
        input_token_size = len(output.prompt_token_ids)
        output_token_size = len(output.outputs[0].token_ids)
        profiler_recv.res(request_id).metric("recvTokenSize", input_token_size).event("httpRes")
        profiler_reply.res(request_id).metric("replyTokenSize", output_token_size).event("httpRes")
    return original_func(this, output, output_type)



# !/usr/bin/python3.7
# -*- coding: utf-8 -*-
# Copyright (c) Huawei Technologies Co., Ltd. 2023-2024. All rights reserved.
"""
1. 训练基础模型
2. 用基础模型拟合概率
3. 用新增加数据继续训练模型
"""
import argparse
import glob
import shutil
from dataclasses import asdict, dataclass
from math import ceil
from pathlib import Path
from typing import Tuple, Optional, List, Union

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from loguru import logger
from pandas import DataFrame
from sklearn.metrics import r2_score, mean_absolute_percentage_error
from sklearn.model_selection import train_test_split
from msserviceprofiler.msguard import Rule
from msserviceprofiler.modelevalstate.analysis import AnalysisState
from msserviceprofiler.modelevalstate.common import _DECODE, _PREFILL, State
from msserviceprofiler.modelevalstate.common import computer_speed_with_second, get_train_sub_path, \
    update_global_coefficient

try:
    from msserviceprofiler.modelevalstate.data_feature.dataset_with_modin import MyDataSetWithModin as MyDataSet
except ModuleNotFoundError:
    try:
        from msserviceprofiler.modelevalstate.data_feature.dataset_with_swifter import MyDataSetWithSwifter as MyDataSet
    except ModuleNotFoundError:
        from msserviceprofiler.modelevalstate.data_feature.dataset import MyDataSet
from msserviceprofiler.modelevalstate.data_feature.dataset import CustomOneHotEncoder, CustomLabelEncoder, \
    preset_category_data
from msserviceprofiler.modelevalstate.data_feature.v1 import FileReader
from msserviceprofiler.modelevalstate.inference.common import HistInfo, model_op_size, OP_EXPECTED_FIELD_MAPPING, \
    OP_SCALE_HIST_FIELD_MAPPING
from msserviceprofiler.modelevalstate.inference.constant import OpAlgorithm
from msserviceprofiler.modelevalstate.model.xgb_state_model import StateXgbModel
from msserviceprofiler.modelevalstate.train.state_param import StateParam
from msserviceprofiler.modelevalstate.inference.utils import save_dataframe_to_csv


@dataclass
class NodeInfo:
    stage: str  # 当前模型状态的类型 Prefill/Decode
    batch_size: int  # 当前状态处理的请求个数


class PretrainModel:

    def __init__(self, state_param: Optional[StateParam] = None, dataset: Optional[MyDataSet] = None,
                 model: Optional[StateXgbModel] = None, plt_data: bool = False):
        self.state_param = state_param
        self.dataset = dataset
        self.model = model
        self.plt_data = plt_data
        self.computer_up_expectations = []
        self.computer_ud_expectations = []
        self.real_up_expectations = []
        self.real_ud_expectations = []
        self.rmse = []
        self.r2 = []
        self.mape = []

    @staticmethod
    def get_decode_and_prefill_time(target: Tuple[NodeInfo], field: str):
        # 获取不同类型batch的运行时间
        _inner_prefill_time = {}
        _inner_decode_time = {}
        for line in target:
            _cur_state = State()
            if line.stage == _PREFILL:
                _cur_state.batch_prefill = line.batch_size
                update_global_coefficient(_inner_prefill_time, _cur_state, getattr(line, field))
            elif line.stage == _DECODE:
                _cur_state.batch_decode = line.batch_size
                update_global_coefficient(_inner_decode_time, _cur_state, getattr(line, field))
        return _inner_prefill_time, _inner_decode_time

    @staticmethod
    def get_up_ud(target: Tuple[NodeInfo], field: str):
        # 获取不同类型的batch运行的单位时延迟
        _inner_up = {}
        _inner_ud = {}
        for line in target:
            _cur_state = State()
            if line.stage == _PREFILL:
                _cur_state.batch_prefill = line.batch_size
                update_global_coefficient(_inner_up, _cur_state, computer_speed_with_second(line, field))
            elif line.stage == _DECODE:
                _cur_state.batch_decode = line.batch_size
                update_global_coefficient(_inner_ud, _cur_state, computer_speed_with_second(line, field))
        return _inner_up, _inner_ud

    @staticmethod
    def get_stage_after_preprocess(row: pd.Series, encoder: Union[CustomOneHotEncoder, CustomLabelEncoder]):
        # 根据预处理后的数据识别该行数据是decode 还是prefill
        if isinstance(encoder) == CustomOneHotEncoder:
            batch_stage_encoder = \
                [encoder.one_hot_encoders[i] for i, v in enumerate(encoder.one_hots) if v.name == "batch_stage"][0]
            _batch_index = [i for i in row.index if "batch_stage" in i]
            stage = batch_stage_encoder.inverse_transform([[int(row[i]) for i in _batch_index]])

        else:
            batch_stage_encoder = \
                [encoder.category_encoders[i] for i, v in enumerate(encoder.category_info) if v.name == "batch_stage"][
                    0]
            stage = batch_stage_encoder.inverse_transform([int(row.batch_stage)])[0]

        return stage

    @staticmethod
    def get_nodes_with_origin_data(features: DataFrame, labels: DataFrame, predict_field: str,
                                   encoder: Union[CustomOneHotEncoder, CustomLabelEncoder]):
        # 获取原来的node信息
        target_data = []
        for ind, row in features.iterrows():
            stage = PretrainModel.get_stage_after_preprocess(row, encoder)
            _cur_node = NodeInfo(stage, row.batch_size)
            setattr(_cur_node, predict_field, labels.iloc[ind, 0])
            target_data.append(_cur_node)
        return tuple(target_data)

    def train(self, lines_data: Optional[DataFrame] = None,
              middle_save_path: Optional[Path] = None):
        logger.info("start train")
        self.dataset.construct_data(lines_data, plt_data=self.plt_data, middle_save_path=middle_save_path)
        rmse = self.model.train(self.dataset, middle_save_path=middle_save_path)
        logger.info(f"rmse {rmse}")

        self.rmse.append(rmse)

    def partial_train(self, lines_data: Optional[DataFrame] = None,
                      middle_save_path: Optional[Path] = None):
        logger.info("start partial train")
        self.dataset.custom_encoder.fit(load=True)
        self.dataset.construct_data(lines_data, plt_data=self.plt_data, middle_save_path=middle_save_path)
        rmse = self.model.train(self.dataset, train_type="partial_fit", middle_save_path=middle_save_path)
        self.rmse.append(rmse)

    def get_nodes_with_model_predict(self, features: DataFrame):
        # 使用模型进行预测
        target_data = []
        for _, row in features.iterrows():
            try:
                _predict = self.model.predict((row,))[0]
                stage = self.get_stage_after_preprocess(row, self.dataset.custom_encoder)
                _cur_node = NodeInfo(stage, row.batch_size)
                setattr(_cur_node, self.state_param.predict_field, _predict)
                target_data.append(_cur_node)
            except Exception as e:
                logger.error(f"get_nodes_with_model_predict时出错：{e}")  
        return tuple(target_data)

    def predict_and_plot(self, features: DataFrame, labels: DataFrame, predict_field: str, save_path: Optional[Path]):
        origin_data = self.get_nodes_with_origin_data(features, labels, predict_field, self.dataset.custom_encoder)
        data = self.get_nodes_with_model_predict(features)
        r2 = r2_score([getattr(k, predict_field) for k in origin_data], [getattr(k, predict_field) for k in data])
        self.r2.append(r2)
        mape = mean_absolute_percentage_error([getattr(k, predict_field) for k in origin_data],
                                              [getattr(k, predict_field) for k in data])
        self.mape.append(mape)
        all_up, all_ud = self.get_up_ud(data, predict_field)
        origin_up, origin_ud = self.get_up_ud(tuple(origin_data), predict_field)

        if self.state_param.plot_velocity_std:
            self.plot_velocity_std(origin_up, all_up, origin_ud, all_ud, save_path=save_path)
        if self.state_param.plot_input_time_with_predict:
            # 绘制时间
            _all_prefill_time, _all_decode_time = self.get_decode_and_prefill_time(data, predict_field)
            origin_prefill_time, origin_decode_time = self.get_decode_and_prefill_time(tuple(origin_data),
                                                                                       predict_field)
            AnalysisState.plot_input_velocity_with_predict(origin_prefill_time, _all_prefill_time, "batch_prefill",
                                                           f"origin and predict prefill time {predict_field} std",
                                                           "batch_prefill",
                                                           "time us", save_path=save_path)
            AnalysisState.plot_input_velocity_with_predict(origin_decode_time, _all_decode_time, "batch_decode",
                                                           f"origin and predict decode time {predict_field} std",
                                                           "batch_decode",
                                                           "time us", save_path=save_path)
        return all_up, all_ud

    def predict_and_plot_with_speed(self, features: DataFrame, labels: DataFrame, save_path: Optional[Path]):
        logger.info("predict test data.")
        _predicts = self.model.predict(features.values)
        _origin_data = labels
        r2 = r2_score(labels, _predicts)
        self.r2.append(r2)
        logger.info(f"r2: {r2}")
        mape = mean_absolute_percentage_error(labels, _predicts)
        self.mape.append(mape)
        logger.info(f"mape: {mape}")
        _predict_df = pd.DataFrame({"batch_stage": self.dataset.load_data["batch_stage"],
                                    "batch_size": self.dataset.load_data["batch_size"],
                                    "predict": _predicts})
        _origin_df = pd.DataFrame({"batch_stage": self.dataset.load_data["batch_stage"],
                                   "batch_size": self.dataset.load_data["batch_size"],
                                   "predict": _predicts})
        if self.state_param.plot_input_time_with_predict:
            AnalysisState.plot_input_velocity_with_df(_predict_df, _origin_df, save_path)

    def predict(self, lines_data: DataFrame,
                save_path: Optional[Path] = None):
        logger.info("start predict")
        self.dataset.construct_data(lines_data, plt_data=self.plt_data, middle_save_path=save_path)
        try:
            return self.predict_and_plot_with_speed(self.dataset.features, self.dataset.labels, save_path)
        except (AttributeError, OverflowError, KeyError, ValueError, RuntimeError):
            return self.predict_and_plot(self.dataset.features, self.dataset.labels, self.state_param.predict_field,
                                         save_path=save_path)

    def plot_velocity_std(self, origin_up, all_up, origin_ud, all_ud, save_path: Optional[Path] = None):
        logger.info("start plot velocity std")
        # 对比Up,Ud的分布
        AnalysisState.plot_input_velocity_with_predict(origin_up, all_up, "batch_prefill",
                                                       f"origin and predict up {self.state_param.predict_field} std",
                                                       "batch_prefill",
                                                       "velocity", save_path=save_path)
        AnalysisState.plot_input_velocity_with_predict(origin_ud, all_ud, "batch_decode",
                                                       f"origin and predict ud {self.state_param.predict_field} std",
                                                       "batch_decode",
                                                       "velocity", save_path=save_path)
        AnalysisState.plot_input_velocity(origin_up, "batch_prefill", f"up {self.state_param.predict_field} std",
                                          "batch_prefill",
                                          "velocity", save_path=save_path)
        AnalysisState.plot_input_velocity(origin_ud, "batch_decode", f"ud {self.state_param.predict_field} std",
                                          "batch_decode",
                                          "velocity", save_path=save_path)

    def bak_model(self, increment_stage: str = "base"):
        _bak_dir = self.state_param.bak_dir.joinpath(increment_stage)
        _bak_dir.mkdir(parents=True, exist_ok=True, mode=0o750)
        shutil.copy(self.state_param.xgb_model_save_model_path,
                    _bak_dir.joinpath(self.state_param.xgb_model_save_model_path.name))

    def plot_metric(self, save_path: Optional[Path] = None):
        data = {"rmse": self.rmse, "r2": self.r2, "mape": self.mape}
        df = pd.DataFrame(data)
        sns.scatterplot(df)
        if save_path:
            plt.savefig(save_path.joinpath("metric.png"))
            plt.close()
        else:
            plt.show()


class TrainVersion1:
    @staticmethod 
    def simple_train(file_paths: List[Path], sp: StateParam, pm: PretrainModel):
        # 训练模型，将全部数据1:9分，9进行训练，1进行预测。
        fl = FileReader(file_paths)
        line_data = fl.read_lines()
        line_data = line_data.dropna()

        def replace_none(value):
            if isinstance(value, str) and 'None' in value:
                return None
            return value
        
        line_data = line_data.applymap(replace_none)
        line_data = line_data.dropna()
        train_data, test_data = train_test_split(line_data, test_size=0.1, shuffle=True)
        logger.info(f"train data shape {train_data.shape}")
        sp.comments = f"input files: {file_paths} \n"
        save_path = sp.step_dir.joinpath("base")
        save_path.mkdir(parents=True, exist_ok=True, mode=0o750)
        pm.train(train_data.reset_index(drop=True), middle_save_path=save_path)
        sp.comments += f'feature shape {pm.dataset.features.shape}\n'
        sp.comments += (f"data shuffle: True, \n train case: {pm.dataset.train_x.shape}, "
                        f"validate case: {pm.dataset.test_x.shape}, predict case: {test_data.shape} \n")
        pm.bak_model()
        logger.info("test data {test_data.shape}")
        save_path = sp.step_dir.joinpath("1")
        save_path.mkdir(parents=True, exist_ok=True, mode=0o750)
        pm.predict(test_data.reset_index(drop=True), save_path)
        logger.info("finished train")


parser = argparse.ArgumentParser(prog="Train Model.")
parser.add_argument("-i", "--input", default=None, type=Path, required=True)
parser.add_argument("-o", "--output", default=Path("output"), type=Path)


def pretrain(input_path, output_path):
    _input_file = Path(input_path).expanduser().resolve()
    if not Rule.input_dir_traverse.is_satisfied_by(_input_file):
        logger.error("not found dir for train model")
        return
    # 获取目录中所有的feature.csv
    train_files = glob.glob(f"{_input_file}/**/*feature.csv", recursive=True)
    if not train_files:
        raise FileNotFoundError(f"{_input_file}/**/*feature.csv")
    train_files = [Path(p) for p in train_files]
    # 创建输出目录
    output = Path(output_path).expanduser().resolve()
    if not output.exists:
        output.mkdir(parents=True, mode=0o750)
    # 运行模型训练
    sp = StateParam(
        base_path=output,
        predict_field="model_execute_time",
        save_model=True,
        shuffle=True,
        plot_pred_and_real=False,
        plot_data_feature=False,
        plot_velocity_std=False,
        plot_predict_std=False,
        op_algorithm=OpAlgorithm.EXPECTED,
        xgb_model_show_test_data_prediction=False,
        xgb_model_show_feature_importance=False,
        plot_input_time_with_predict=False,
        title="MixModel without warmup with service info"
    )
    model = StateXgbModel(
        train_param=sp.xgb_model_train_param,
        update_param=sp.xgb_model_update_param,
        save_model_path=sp.xgb_model_save_model_path,
        load_model_path=sp.xgb_model_save_model_path,
        show_test_data_prediction=sp.xgb_model_show_test_data_prediction,
        show_feature_importance=sp.xgb_model_show_feature_importance,
    )
    custom_encoder = CustomLabelEncoder(preset_category_data)
    custom_encoder.fit()
    dataset = MyDataSet(custom_encoder=custom_encoder, predict_field=sp.predict_field,
                        shuffle=sp.shuffle, op_algorithm=sp.op_algorithm)
    pm = PretrainModel(state_param=sp, dataset=dataset, model=model, plt_data=sp.plot_data_feature)
    TrainVersion1.simple_train(train_files, sp, pm)
    train_data = dataset.features.copy(deep=False)
    train_data["label"] = dataset.labels
    _train_dir = output.joinpath("cache")
    save_dataframe_to_csv(train_data, _train_dir, "train_data.csv")


def main(args):
    # 解析命令
    _input_file = Path(args.input).expanduser().resolve()
    output = Path(args.output).expanduser().resolve()
    pretrain(_input_file, output)





# -*- coding: utf-8 -*-
# Copyright (c) Huawei Technologies Co., Ltd. 2025-2025. All rights reserved.
import os
import shutil
from pathlib import Path
from typing import Optional
from loguru import logger
from pydantic import BaseModel, Field
from msserviceprofiler.msguard import Rule


class AisBenchCommandConfig(BaseModel):
    models: str = ""
    datasets: str = ""
    mode: str = ""
    num_prompts: str = ""
    work_dir: str = ""


class AisBenchCommand:
    def __init__(self, aisbench_command_config: AisBenchCommandConfig):
        self.process = shutil.which("ais_bench")
        if self.process is None:
            raise ValueError("Error: The 'ais_bench' executable was not found in the system PATH.")
        self.aisbench_command_config = aisbench_command_config

    @property
    def command(self):
        _cmd = [self.process,
                "--models", self.aisbench_command_config.models,
                "--datasets", self.aisbench_command_config.datasets,
                "--mode", self.aisbench_command_config.mode,
                "--num-prompts", self.aisbench_command_config.num_prompts,
                "--work-dir", self.aisbench_command_config.work_dir,
                "--debug"
                ]
        return _cmd


class BenchmarkCommandConfig(BaseModel):
    dataset_path: str = ""
    dataset_type: str = "gsm8k"
    model_name: str = ""
    model_path: str = ""
    test_type: str = "client"
    max_output_len: str = ""
    http: str = ""
    management_http: str = ""
    warmup_size: str = "1"
    tokenizer: str = "True"
    save_path: str = ""
    request_num: str = ""
    request_count: str = ""

 
class BenchmarkCommand:
    def __init__(self, benchmark_command_config: BenchmarkCommandConfig):
        self.process = shutil.which("benchmark")
        if self.process is None:
            raise ValueError("Error: The 'benchmark' executable was not found in the system PATH.")
        self.benchmark_command_config = benchmark_command_config
 
    @property
    def command(self):
        if not Rule.input_file_read.is_satisfied_by(self.benchmark_command_config.dataset_path):
            logger.error("the file of dataset_path is not safe, please check")
            return None
        
        _cmd = [self.process,
                "--DatasetPath", self.benchmark_command_config.dataset_path,
                "--DatasetType", self.benchmark_command_config.dataset_type,
                "--ModelName", self.benchmark_command_config.model_name,
                "--ModelPath", self.benchmark_command_config.model_path,
                "--TestType", self.benchmark_command_config.test_type,
                "--MaxOutputLen", self.benchmark_command_config.max_output_len,
                "--Http", self.benchmark_command_config.http,
                "--ManagementHttp", self.benchmark_command_config.management_http,
                "--Concurrency", "$CONCURRENCY",
                "--RequestRate", "$REQUESTRATE",
                "--WarmupSize", self.benchmark_command_config.warmup_size,
                "--Tokenizer", self.benchmark_command_config.tokenizer,
                "--SavePath", self.benchmark_command_config.save_path,
                ]
        if self.benchmark_command_config.request_num:
            _cmd.extend(["--RequestNum", self.benchmark_command_config.request_num])
        if self.benchmark_command_config.request_count:
            _cmd.extend(["--RequestCount", self.benchmark_command_config.request_count])
        return _cmd
 
 
class VllmBenchmarkCommandConfig(BaseModel):
    serving: str = ""
    backend: str = "vllm"
    host: str = ""
    port: str = ""
    model: str = ""
    served_model_name: str = ""
    dataset_name: str = ""
    dataset_path: str = ""
    num_prompts: str = ""
    result_dir: str = ""
    others: str = ""

 
 
class VllmBenchmarkCommand:
    def __init__(self, benchmark_command_config: VllmBenchmarkCommandConfig):
        self.benchmark_command_config = benchmark_command_config
 
    @property
    def command(self):
        if not Rule.input_file_read.is_satisfied_by(self.benchmark_command_config.dataset_path):
            logger.error("the file of dataset_path is not safe, please check")
            return None
        cmd = ["python", self.benchmark_command_config.serving,
                "--backend", self.benchmark_command_config.backend,
                "--host", self.benchmark_command_config.host,
                "--port", self.benchmark_command_config.port,
                "--model", self.benchmark_command_config.model,
                "--served-model-name", self.benchmark_command_config.served_model_name,
                "--dataset-name", self.benchmark_command_config.dataset_name,
                "--dataset-path", self.benchmark_command_config.dataset_path,
                "--num-prompts", self.benchmark_command_config.num_prompts,
                "--max-concurrency", "$CONCURRENCY",
                "--request-rate", "$REQUESTRATE",
                "--result-dir", "$MODEL_EVAL_STATE_VLLM_CUSTOM_OUTPUT",
                "--save-result"]
        if self.benchmark_command_config.others:
            cmd.extend(self.benchmark_command_config.others.split())
        return cmd
 
 
class MindieCommandConfig(BaseModel):
    pass
 
 
class MindieCommand:
    def __init__(self, command_config: MindieCommandConfig):
        self.command_config = command_config
 
    @property
    def command(self):
        mindie_service_default_path: str = "/usr/local/Ascend/mindie/latest/mindie-service"
        mindie_service_path: str = os.getenv("MIES_INSTALL_PATH", mindie_service_default_path)
        mindie_command_path: str = os.path.join(mindie_service_path, "bin", "mindieservice_daemon")
        return [mindie_command_path]
 

class KubectlCommandConfig(BaseModel):
    kubectl_default_path: Path = Path("")
    kubectl_single_path: Optional[Path] = Field(
        default_factory=lambda data: data["kubectl_default_path"].joinpath("deploy.sh").resolve())
    kubectl_log_path: Optional[Path] = Field(
        default_factory=lambda data: data["kubectl_default_path"].joinpath("show_logs.sh").resolve())


class KubectlCommand():
    def __init__(self, command_config: KubectlCommandConfig):
        self.command_config = command_config
    
    @property
    def command(self):
        kubectl_command_path = self.command_config.kubectl_single_path
        cmd = ['bash', kubectl_command_path]
        return cmd

    @property
    def log_command(self):
        kubectl_path = shutil.which("kubectl")
        cmd = [kubectl_path, "get", "pods", "-A", "-owide"]
        return cmd

    @property
    def monitor_command(self):
        kubectl_path = shutil.which("kubectl")
        cmd = [kubectl_path, "logs", "-f", "-n", "mindie"]
        return cmd


class VllmCommandConfig(BaseModel):
    host: str = ""
    port: str = ""
    model: str = ""
    served_model_name: str = ""
    others: str = ""
 
 
class VllmCommand:
    def __init__(self, command_config: VllmCommandConfig):
        self.process = shutil.which("vllm")
        if self.process is None:
            raise ValueError("Error: The 'vllm' executable was not found in the system PATH.")
        self.command_config = command_config
 
    @property
    def command(self):
        cmd = [self.process, "serve",
                self.command_config.model,
                "--served-model-name", self.command_config.served_model_name,
                "--host", self.command_config.host,
                "--port", self.command_config.port,
                "--max-num-batched-tokens", "$MAX_NUM_BATCHED_TOKENS",
                "--max-num-seqs", "$MAX_NUM_SEQS"]
        if self.command_config.others:
            cmd.extend(self.command_config.others.split())
        return cmd


# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import csv
from pathlib import Path
from typing import Dict, Optional, List, Tuple

import numpy as np
import pandas as pd
from loguru import logger
from msserviceprofiler.msguard.security import open_s, sanitize_csv_value
from msserviceprofiler.modelevalstate.config.config import (
    DataStorageConfig,
    RUN_TIME,
    PerformanceIndex,
    OptimizerConfigField,
    settings
)
from msserviceprofiler.modelevalstate.optimizer.benchmark import BenchMark, VllmBenchMark, AisBench
from msserviceprofiler.modelevalstate.optimizer.simulator import Simulator, VllmSimulator
from msserviceprofiler.modelevalstate.common import read_csv_s


LLM_MODEL = "llm_model"
DATASET_PATH = "dataset_path"
SIMULATOR = "simulator"
NUM_PROMPTS = "num_prompts"
MAX_OUTPUT_LEN = "max_output_len"


class DataStorage:
    def __init__(self, config: DataStorageConfig, simulator=None, benchmark=None, ):
        self.config = config
        if not self.config.store_dir.exists():
            self.config.store_dir.mkdir(parents=True, mode=0o750)
        self.save_file = self.config.store_dir.joinpath(f"data_storage_{RUN_TIME}.csv")
        self.simulator = simulator
        self.benchmark = benchmark

    @staticmethod
    def load_history_position(load_dir: Path, filter_field: Optional[Dict] = None) -> Optional[List]:
        if not load_dir.exists():
            raise FileNotFoundError(f"file: {load_dir}")
        if not load_dir.is_dir():
            raise ValueError(f"Expect a directory, not a file.")
        history_data = []
        for file in sorted([f for f in load_dir.iterdir() if f.is_file()], key=lambda x: x.stat().st_ctime):
            if file.name.startswith("data_storage") and file.suffix == ".csv":
                data = read_csv_s(file).to_dict(orient="records")
                history_data.extend(data)
        if not history_data:
            return None
        return DataStorage.filter_data(history_data, filter_field)

    @staticmethod
    def filter_data(datas: List[Dict], filter_field: Optional[Dict] = None):
        if not filter_field:
            return datas
        filter_datas = []
        for d in datas:
            flag = False
            for k, v in filter_field.items():
                # 不存在的字段 无法进行筛选
                if k not in d:
                    continue
                # 字段存在 但不等于目标字段，则去掉
                if d[k].strip().lower() != v.strip().lower():
                    flag = True
                    break
            if flag:
                continue
            else:
                filter_datas.append(d)
        return filter_datas

    def get_run_info(self):
        _run_info = {}
        if isinstance(self.benchmark, BenchMark):
            _run_info[NUM_PROMPTS] = self.benchmark.benchmark_config.command.request_count
        elif isinstance(self.benchmark, (VllmBenchMark, AisBench)):
            _run_info[NUM_PROMPTS] = self.benchmark.benchmark_config.command.num_prompts
        return _run_info

    def save(self, performance_index: PerformanceIndex, params: Tuple[OptimizerConfigField], **kwargs):
        logger.info(f"Save result with DataStorage. File path: {self.save_file}")
        _column = []
        _value = []
        for k, v in performance_index.model_dump().items():
            _column.append(k)
            _value.append(v)
        for _p in params:
            _column.append(_p.name)
            _value.append(_p.value)
        for k, v in kwargs.items():
            _column.append(k)
            _value.append(v)
        for k, v in self.get_run_info().items():
            _column.append(k)
            _value.append(v)
        if self.save_file.exists():
            with open_s(self.save_file, "a+") as f:
                data_writer = csv.writer(f)
                data_writer.writerow([sanitize_csv_value(_v) for _v in _value])
        else:
            with open_s(self.save_file, "w") as f:
                data_writer = csv.writer(f)
                data_writer.writerow(_column)
                data_writer.writerow([sanitize_csv_value(_v) for _v in _value])

    def get_best_result(self):
        optimizer_result = read_csv_s(self.save_file)
        optimizer_result = optimizer_result.replace([np.inf, -np.inf], np.nan)
        pso_result = optimizer_result
        if isinstance(self.benchmark, (BenchMark, VllmBenchMark, AisBench)):
            # 提取公共属性访问
            command = self.benchmark.benchmark_config.command
            request_nums = command.request_count if isinstance(self.benchmark, BenchMark) else command.num_prompts
            if isinstance(request_nums, str) and request_nums.isdigit():
                pso_result = optimizer_result[optimizer_result[NUM_PROMPTS] == int(request_nums)]
        pso_result = pso_result.dropna(subset="fitness")
        pso_result = pso_result[pso_result["time_to_first_token"] > 0]
        pso_result = pso_result[pso_result["time_per_output_token"] > 0]
        pso_result = pso_result[pso_result["generate_speed"] > 0]
        pso_result = pso_result.reset_index()
        _fitness_index = pso_result.nsmallest(self.config.pso_top_k, "fitness").index
        if settings.ttft_penalty and settings.tpot_penalty:
            _generate_speed_index = pso_result[
                (pso_result["time_to_first_token"] <= settings.ttft_slo * (1 + settings.slo_coefficient)) &
                (pso_result["time_per_output_token"] <= settings.tpot_slo * (1 + settings.slo_coefficient))].nlargest(
                self.config.pso_top_k, "generate_speed").index
        elif settings.tpot_penalty:
            _generate_speed_index = pso_result[
                pso_result["time_per_output_token"] <= settings.tpot_slo * (1 + settings.slo_coefficient)].nlargest(
                self.config.pso_top_k, "generate_speed").index
        else:
            _generate_speed_index = pso_result.nlargest(self.config.pso_top_k, "generate_speed").index
        _fine_tune_index = _fitness_index.union(_generate_speed_index)
        return pso_result.iloc[_fine_tune_index]


# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ms_service_profiler import Profiler, Level
from ..module_hook import vllm_hook


@vllm_hook(("vllm.engine.async_llm_engine", "AsyncLLMEngine.add_request"), min_version="0.9.1")
async def add_request_async(original_func, this, request_id, prompt, *args, **kwargs):
    Profiler(Level.INFO).domain("Request").res(request_id).event("httpReq")
    Profiler(Level.INFO).domain("Request").res(request_id).event("tokenize")
    return await original_func(this, request_id, prompt, *args, **kwargs)


@vllm_hook(("vllm.v1.engine.output_processor", "OutputProcessor.process_outputs"), min_version="0.9.1")
def process_outputs(original_func, this, engine_core_outputs, *args, **kwargs):
    if len(engine_core_outputs) == 0:
        return original_func(this, engine_core_outputs, *args, **kwargs)

    request_id_list = []
    for engine_core_output in engine_core_outputs:
        request_id = engine_core_output.request_id
        request_state = this.request_states.get(request_id)
        request_id_list.append(request_id)
        if request_state and engine_core_output.finish_reason is not None:
            recv_token_size = len(request_state.prompt_token_ids)
            reply_token_size = (request_state.stats.num_generation_tokens if request_state.stats else None)
            
            profiler = Profiler(Level.INFO).domain("Request").res(request_id)
            profiler = profiler.\
                metric("recvTokenSize", recv_token_size).\
                metric("replyTokenSize", reply_token_size)
            profiler.event("httpRes")

    ret = original_func(this, engine_core_outputs, *args, **kwargs)
    prof = Profiler(Level.INFO).domain("Request").res(request_id_list)
    prof.event("detokenize")
    return ret



# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import Counter
from ms_service_profiler import Profiler, Level
from ..module_hook import vllm_hook
from ..utils import logger


@vllm_hook(("vllm.v1.core.kv_cache_manager", "KVCacheManager.allocate_slots"), min_version="0.9.1")
def allocate_slots(original_func, this, request, *args, **kwargs):
    ret = original_func(this, request, *args, **kwargs)
    num_blocks = this.block_pool.get_num_free_blocks()
    prof = Profiler(Level.INFO).domain("KVCache").res(request.request_id)
    prof.metric("deviceBlock", num_blocks).event("Allocate")
    return ret


@vllm_hook(("vllm.v1.core.kv_cache_manager", "KVCacheManager.free"), min_version="0.9.1")
def free(original_func, this, request, *args, **kwargs):
    ret = original_func(this, request, *args, **kwargs)
    num_blocks = this.block_pool.get_num_free_blocks()
    prof = Profiler(Level.INFO).domain("KVCache").res(request.request_id)
    prof.metric("deviceBlock", num_blocks).event("Free")
    return ret


@vllm_hook(("vllm.v1.core.kv_cache_manager", "KVCacheManager.get_computed_blocks"), min_version="0.9.1")
def get_computed_blocks(original_func, this, request, *args, **kwargs):
    ret = original_func(this, request, *args, **kwargs)
    if len(ret) > 1 and request.num_tokens > 0:
        num_new_computed_tokens = ret[1]
        cur_hit_rate = num_new_computed_tokens / request.num_tokens
        prof = Profiler(Level.INFO).domain("KVCache").res(request.request_id)
        prof.attr("hitCache", cur_hit_rate).event("GetCacheHitRate")
    return ret



# !/usr/bin/python3.7
# -*- coding: utf-8 -*-
# Copyright (c) Huawei Technologies Co., Ltd. 2023-2024. All rights reserved.

import os
import subprocess
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Dict
import shutil
import pandas as pd
from loguru import logger
from msserviceprofiler.msguard import validate_params, Rule


_PREFILL = "prefill"
_DECODE = "decode"
_KEY_WORD = "H" + "B" + "M"


class StateType(Enum):
    DEFAULT = 0
    LINE = 1


@dataclass
class State:
    prefill: int = 0
    decode: int = 0
    batch_prefill: int = 0
    batch_decode: int = 0

    def __repr__(self):
        return f"TT_{self.prefill}_{self.decode}_{self.batch_prefill}_{self.batch_decode}"

    def __hash__(self):
        return hash(self.__repr__())

    def __eq__(self, other):
        return self.__repr__() == other.__repr__()

    def __ne__(self, other):
        return self.__repr__() != other.__repr__()

    def sum(self):
        return self.prefill + self.decode + self.batch_prefill + self.batch_decode


def computer_speed(line_node, field):
    if getattr(line_node, field) == 0:
        return 1
    return 1 / (getattr(line_node, field) * 10 ** -3)


def computer_speed_with_second(line_node, field):
    if getattr(line_node, field) == 0:
        return 1
    return 1 / (getattr(line_node, field) * 10 ** -6)


def get_train_sub_path(base_path: Path):
    # 给训练输出目录生成新的目录
    if not base_path.exists():
        base_path.mkdir(parents=True, exist_ok=True, mode=0o750)
    _sub_len = len([0 for _ in base_path.iterdir()])
    _sub_dir = base_path.joinpath(f"{_sub_len + 1}")
    _sub_dir.mkdir(parents=True, exist_ok=True, mode=0o750)
    return _sub_dir


def update_global_coefficient(global_coefficient: Dict, key: State, value: float) -> None:
    if key not in global_coefficient:
        global_coefficient[key] = [value]
    else:
        global_coefficient[key].append(value)


def get_module_version(module_name):
    try:
        # 方法1：直接导入模块
        import importlib
        module = importlib.import_module(module_name)
        if hasattr(module, "__version__"):
            return module.__version__
        elif hasattr(module, "version"):
            return module.version
    except ImportError:
        pass

    try:
        # 方法2：使用 pkg_resources
        import pkg_resources
        return pkg_resources.get_distribution(module_name).version
    except (ImportError, pkg_resources.DistributionNotFound):
        pass

    try:
        # 方法3：使用 importlib.metadata（Python 3.8+）
        import importlib.metadata
        return importlib.metadata.version(module_name)
    except (ImportError):
        pass

    # # 方法4：最后尝试 pip show
    _flag = "MODEL_EVAL_STATE_GET_MODULE_VERSION_FLAG"
    try:
        if os.getenv(_flag) == "true":
            raise ValueError
        pip_path = shutil.which("pip3")
        if pip_path is not None:
            output = subprocess.check_output(
                [pip_path, "show", module_name],
                universal_newlines=True,
                env={"MODEL_EVAL_STATE_GET_MODULE_VERSION_FLAG": "true"}
            )
            for line in output.splitlines():
                if line.startswith("Version:"):
                    return line.split(":")[1].strip()
    except (subprocess.CalledProcessError, FileNotFoundError):
        pass

    raise ValueError("模块未安装或无法获取版本")


@validate_params({"path": Rule.input_file_read})
def read_csv_s(path, **kwargs):
    try:
        return pd.read_csv(path, **kwargs)
    except Exception as e:
        raise ValueError(f"Failed to read csv %r." % path) from e


def is_mindie():
    try:
        import mindie_llm
    except ModuleNotFoundError:
        return False
    return True


def is_vllm():
    try:
        import vllm
    except ModuleNotFoundError:
        return False
    return True


def ais_bench_exists():
    try:
        import ais_bench
    except ModuleNotFoundError:
        return False
    return True


def get_npu_total_memory(device_id: int = 0):
    cmd = ["npu-smi", "info", "-t", "usages", "-i", str(device_id)]
    memory_key_word = _KEY_WORD + " Capacity(MB)"
    usage_rate_key_word = _KEY_WORD + " Usage Rate(%)"
    try:
        output = subprocess.check_output(cmd)
        output = output.decode("utf-8")
        total_memory_line = [line for line in output.splitlines() if memory_key_word in line][0]
        total_memory_line = total_memory_line.split(":")[1].strip()
        memory_usage_rate = [line for line in output.splitlines() if usage_rate_key_word in line][0]
        memory_usage_rate = memory_usage_rate.split(":")[1].strip()
        memory_line, usage_rate = int(total_memory_line), int(memory_usage_rate)
        logger.debug(f"cmd: {cmd}, result: {memory_line}, {usage_rate}")
        return memory_line, usage_rate
    except Exception as e:
        raise ValueError(f"Failed to retrieve total video memory. Please check if the video memory query command {cmd} "
                     f"matches the current parsing code. ") from e



# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import atexit
import json
import os
import queue
import signal
import stat
import threading
import time
from pathlib import Path
from threading import Thread
from typing import Optional
from loguru import logger
import numpy as np
import torch
from loguru import logger

from msserviceprofiler.modelevalstate.config.config import settings
from msserviceprofiler.modelevalstate.inference.constant import IS_SLEEP_FLAG, BatchStage
from msserviceprofiler.modelevalstate.inference.data_format_v1 import BatchField, RequestField, ConfigPath
from msserviceprofiler.modelevalstate.inference.dataset import CustomLabelEncoder, preset_category_data, DataProcessor
from msserviceprofiler.modelevalstate.inference.file_reader import FileHanlder, StaticFile
from msserviceprofiler.modelevalstate.inference.state_eval_v1 import predict_v1_with_cache
from msserviceprofiler.msguard.security import open_s

predict_queue = queue.Queue()


class ServiceField:
    batch_field = None
    request_field = None
    next_tokens = None
    config_path = None
    fh = None
    data_processor = None
    req_id_and_max_decode_length = None


ServiceField.config_path = ConfigPath(settings.latency_model.model_path,
                                      settings.latency_model.static_file_dir,
                                      settings.latency_model.req_and_decode_file,
                                      settings.latency_model.cache_data)

sub_thread: Optional[Thread] = None


class FileLogger:
    def __init__(self, file_path, mode="a"):
        self.file_path = file_path
        self.fout = None
        self.lock = threading.Lock()
        self.mode = mode

    def open_file(self):
        flags = os.O_WRONLY | os.O_CREAT
        modes = stat.S_IWUSR | stat.S_IRUSR
        if not isinstance(self.file_path, Path):
            self.file_path = Path(self.file_path)
        self.fout = os.fdopen(os.open(self.file_path, flags, modes), self.mode, buffering=1024)

    def write(self, message):
        with self.lock:
            if self.fout:
                self.fout.write(message)
                self.fout.write("\n")
                self.fout.flush()

    def close(self):
        with self.lock:
            if self.fout:
                self.fout.close()
                self.fout = None


def write_file(file_logger):
    file_logger.open_file()
    while True:
        flag = False
        items = []
        while not predict_queue.empty():
            items.append(predict_queue.get())
        for res in items:
            if res is None:
                flag = True
                break
            file_logger.write(str(res))
        if flag:
            break
        time.sleep(1)
    file_logger.close()


def signal_handler(file_logger):
    predict_queue.put(None)
    if sub_thread:
        sub_thread.join(timeout=3)
        if sub_thread.is_alive():
            raise TimeoutError("子线程未在指定时间完成")
    file_logger.close()


file_log = FileLogger(Path(settings.simulator_output).joinpath(f"simulate_{os.getpid()}.csv"))
atexit.register(signal_handler, file_log)


class Simulate:
    first = True
    predict_cache = {}
    request_seq_len = {}

    @staticmethod
    def init(plugin_object):
        if Simulate.first:
            if isinstance(plugin_object.input_manager.cache_config.eos_token_id, int):
                plugin_object.eos_token_id = plugin_object.input_manager.cache_config.eos_token_id
            else:
                plugin_object.eos_token_id = plugin_object.input_manager.cache_config.eos_token_id[0]
            if ServiceField.config_path.req_and_decode_file.exists():
                with open_s(ServiceField.config_path.req_and_decode_file, 'r') as f:
                    try:
                        req_and_decode = json.load(f)
                    except json.JSONDecodeError as e:
                        raise ValueError(f"Invalid JSON in config file: {e}") from e
                    ServiceField.req_id_and_max_decode_length = {
                        int(k): int(v) 
                        for k, v in req_and_decode.items()
                        if str(k).strip().isdigit() and str(v).strip().isdigit()
                    }
            else:
                ServiceField.req_id_and_max_decode_length = {}
            if not Path(ServiceField.config_path.static_file_dir).exists():
                Path(ServiceField.config_path.static_file_dir).mkdir(parents=True, mode=0o750)
            static_file = StaticFile(base_path=ServiceField.config_path.static_file_dir)
            ServiceField.fh = FileHanlder(static_file)
            ServiceField.fh.load_static_data()
            custom_encoder = CustomLabelEncoder(preset_category_data)
            custom_encoder.fit()
            ServiceField.data_processor = DataProcessor(custom_encoder)
            Simulate.first = False
            global sub_thread
            sub_thread = threading.Thread(target=write_file, args=(file_log,))
            sub_thread.start()

    @staticmethod
    def generate_random_token(plugin_object, shape, max_value=32000):
        # max_value 是vacab size，就是词表的范围
        if np.prod(shape) > max_value + 1:
            raise ValueError("token数量超过词表的范围，无法进行无放回抽样")
        array = np.random.choice(np.arange(0, max_value + 1), size=np.prod(shape), replace=False)
        array = np.reshape(array, shape)
        array = np.where(array == plugin_object.eos_token_id, np.random.randint(0, max_value + 1), array)
        return array

    @staticmethod
    def generate_logits(batch_size, vocab_size: int = 129280, device="npu:0", dtype="float16"):
        dtype_map = {
            torch.float16: "float16",
            torch.bfloat16: "bfloat16",
            torch.float: "float",
            torch.int8: "int8"
        }
        _cur_dtype = torch.float16
        for k, v in dtype_map.items():
            if v == dtype:
                _cur_dtype = k
                break
        tensor = torch.randn((batch_size, vocab_size), dtype=_cur_dtype, device=device)
        return tensor

    @staticmethod
    def generate_features(plugin_object, input_metadata, cached_ids):
        output_len_count = plugin_object.input_manager.cache.output_len_count[cached_ids]
        if input_metadata.is_prefill:
            batch_stage = BatchStage.PREFILL
            for i, req_id in enumerate(input_metadata.batch_request_ids):
                Simulate.request_seq_len[req_id] = input_metadata.batch_seq_len[i]
            all_input_length = input_metadata.batch_seq_len
        else:
            batch_stage = BatchStage.DECODE
            all_input_length = []
            for req_id in input_metadata.batch_request_ids:
                all_input_length.append(Simulate.request_seq_len[req_id])
        all_need_blocks = np.count_nonzero(input_metadata.block_tables > -1, axis=-1)
        request_field = []
        _total_req_input_len = []
        try:
            for _cache_id_index, _cache_id in enumerate(cached_ids):
                _req_input_len = all_input_length[_cache_id_index] - output_len_count[_cache_id_index]
                _total_req_input_len.append(_req_input_len)
                request_field.append(RequestField(_req_input_len,
                                                all_need_blocks[_cache_id_index],
                                                output_len_count[_cache_id_index]))
            batch_field = BatchField(batch_stage, input_metadata.batch_size,
                                    np.count_nonzero(input_metadata.block_tables > -1, axis=-1).sum(),
                                    sum(_total_req_input_len), max(_total_req_input_len))
        except Exception as e:
            logger.warning(f"process ServiceField in generate_features failed: {e}")
            ServiceField.batch_field = None
            ServiceField.request_field = None
            batch_field = None
            return batch_field, request_field

        request_field = tuple(sorted(request_field))
        ServiceField.batch_field = batch_field
        ServiceField.request_field = request_field
        return batch_field, request_field

    @staticmethod
    def update_token(plugin_object, input_metadata, cached_ids, sampling_output):
        output_len_count = plugin_object.input_manager.cache.output_len_count[cached_ids]
        for i, token_id in enumerate(sampling_output.token_ids):
            _cur_out_len = output_len_count[i]
            if input_metadata.batch_request_ids[i] not in ServiceField.req_id_and_max_decode_length:
                continue
            _max_out_len = ServiceField.req_id_and_max_decode_length[input_metadata.batch_request_ids[i]]
            if _cur_out_len < (_max_out_len - 1):
                if token_id == plugin_object.eos_token_id:
                    token_id = np.random.randint(0, plugin_object.model_wrapper.config.vocab_size)
                if sampling_output.top_token_ids.size == 0:
                    continue
                if i >= sampling_output.top_token_ids.size:
                    continue

                if sampling_output.top_token_ids[i].any() and plugin_object.eos_token_id in \
                        sampling_output.top_token_ids[i]:
                    sampling_output.top_token_ids[i] = np.where(
                        sampling_output.top_token_ids[i] == plugin_object.eos_token_id,
                        np.random.randint(0, plugin_object.model_wrapper.config.vocab_size),
                        sampling_output.top_token_ids[i])
            else:
                sampling_output.token_ids[i] = plugin_object.eos_token_id
                if sampling_output.top_token_ids.size == 0:
                    continue
                if i >= sampling_output.top_token_ids.size:
                    continue
                if sampling_output.top_token_ids[i].any():
                    sampling_output.top_token_ids[i] = np.where(
                        sampling_output.top_token_ids[i] != plugin_object.eos_token_id,
                        plugin_object.eos_token_id,
                        sampling_output.top_token_ids[i])


    @staticmethod
    def predict(time_sleep: bool = True):
        """

        return: time ms
        """
        time_sleep = os.getenv(IS_SLEEP_FLAG, str(time_sleep)).lower().strip() == "true"
        # 增加缓存
        st = time.perf_counter()
        _cache_key = (ServiceField.batch_field, ServiceField.request_field)
        if _cache_key not in Simulate.predict_cache:
            predict_res = predict_v1_with_cache(ServiceField.batch_field, ServiceField.request_field,
                                                ServiceField.config_path, ServiceField.fh, ServiceField.data_processor,
                                                cache_data=ServiceField.config_path.cache_data)
            Simulate.predict_cache[_cache_key] = predict_res
        else:
            predict_res = Simulate.predict_cache.get(_cache_key)

        for _pre_v in predict_res:
            if _pre_v == -1:
                continue
            if time_sleep:
                _run_time = time.perf_counter() - st
                _wait_time = _pre_v / 10 ** 6 - _run_time
                if _wait_time > 0:
                    time.sleep(_wait_time)
                return _pre_v
            else:
                return _pre_v

    @staticmethod
    def predict_and_save(time_sleep: bool = False):
        res = Simulate.predict(time_sleep)
        # 增加异步写入能力
        try:
            predict_queue.put_nowait(res)
        except queue.Full:
            logger.error("predict_queue is full")
            predict_queue.put(res)


# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
import pandas as pd
from ms_service_profiler.exporters.utils import save_dataframe_to_csv

from ..common.constants import MAX_BATCH_NUMBER
from ..common.utils import logger
from ..common.split_utils import (
    CSV_COLUMNS, RENAMED_COLUMNS, get_statistics_data, PREFILL_NAME
)


class BaseFrameworkProcessor:
    batch_start_name = "batch_start"
    batch_end_name = "batch_end"
    http_start_name = "http_start"
    http_end_name = "http_end"
    key_name = "key"
    all_time_name = "all_time"
    name_list = [batch_start_name, http_start_name, batch_end_name]
    http_list = [http_start_name, http_end_name]
    # 记录不计算中间时间的事件(异步, 时间掩盖)或无需计算
    filter_list = [http_end_name, all_time_name]

    @staticmethod
    def _get_batch_index(full_batch, start_index, framework_df, key_pid, key_tid):
        # 从总表中获取完整的一个batch
        df_list = []
        current_index = start_index
        index_mask = framework_df.index > start_index
        for name in full_batch:
            name_mask = framework_df["name"] == name
            pid_mask = framework_df["pid"] == key_pid
            tid_mask = framework_df["tid"] == key_tid

            conditions = [
                index_mask & name_mask & pid_mask & tid_mask,
                index_mask & name_mask & pid_mask,
                index_mask & name_mask
            ]
            index = None
            for condition in conditions:
                mask = condition
                if mask.any():
                    index = framework_df[mask].index[0]
                    break
            # 找不到的点跳过
            if index is None:
                continue
            
            df_list.append(index)
            current_index = index
            # 更新index_mask
            index_mask = framework_df.index > current_index
            
        if not df_list:
            return pd.DataFrame()

        return framework_df.loc[df_list]
    
    @staticmethod
    def _calc_during_time(filter_df):
        new_rows = []
        # 遍历每个事件对
        for i in range(len(filter_df) - 1):
            current_row = filter_df.iloc[i]
            next_row = filter_df.iloc[i + 1]

            current_name = current_row["name"]
            next_name = next_row["name"]

            current_end = current_row["end_time(ms)"]
            next_start = next_row["start_time(ms)"]

            # 创建新行
            new_row = {
                "name": f"Between-{current_name}-{next_name}",
                "during_time(ms)": next_start - current_end,
                "start_time(ms)": current_end,
                "end_time(ms)": next_start,
                "pid": current_row["pid"],
                "tid": current_row["tid"],
            }
            new_rows.append(new_row)

        new_df = pd.DataFrame(new_rows)

        return new_df

    @classmethod
    @abstractmethod
    def initialize(cls, args):
        cls.args = args

    def run_split(self, framework_df, name):
        framework_df = self.preprocess_framework_df(framework_df)
        if framework_df.empty:
            return
        filter_df = self.get_filter_df(framework_df, name)
        add_all_time_df = self.get_batch_all_time(filter_df, name)
        framework_df = self.process_exporter(add_all_time_df, name)
        lower_name = name.lower()
        if self.args.log_level == "debug":
            save_dataframe_to_csv(add_all_time_df, self.args.output_path, f"{lower_name}_detail.csv")
            save_dataframe_to_csv(framework_df, self.args.output_path, f"{lower_name}_{self.args.batch_num}.csv")
        filter_name = self.http_start_name if name == PREFILL_NAME else self.batch_start_name
        framework_df = get_statistics_data(framework_df, filter_name, name)
        if not framework_df.empty:
            save_dataframe_to_csv(framework_df, self.args.output_path, f"{lower_name}.csv")

    def preprocess_framework_df(self, framework_df):
        try:
            framework_df = framework_df[framework_df["name"].isin(self.name_list)]
            framework_df = framework_df[CSV_COLUMNS]
        except KeyError as e:
            logger.warning(f"Field '{e.args[0]}' not found in datasource.")
            return pd.DataFrame()

        framework_df = framework_df.rename(columns=RENAMED_COLUMNS)
        
        return framework_df
    
    def get_filter_df(self, framework_df, name):
        """
        动态启停场景下 过滤不完整的batch
        """
        filter_name = self.http_start_name if name == PREFILL_NAME else self.batch_start_name

        valid_indices = framework_df["name"] == filter_name
        if not valid_indices.any():
            logger.warning(f"{name}: No data named {filter_name}")
            return framework_df
        
        first_index = framework_df[valid_indices].index[0]

        # 过滤 DataFrame
        return framework_df.loc[first_index:]
    
    def get_batch_all_time(self, framework_df, name):
        batch_rows = framework_df[framework_df["name"] == self.batch_start_name]
        
        if len(batch_rows) < 2:
            logger.warning(f"{name}: The length of {self.batch_start_name} is less two")
            return framework_df
        
        def create_all_time_rows(group):
            new_rows = []
            for i in range(len(group) - 1):
                current_row = group.iloc[i]
                next_row = group.iloc[i + 1]
                during_time = next_row["start_time(ms)"] - current_row["start_time(ms)"]

                all_time_row = {
                    "name": self.all_time_name,
                    "start_time(ms)": current_row["start_time(ms)"],
                    "end_time(ms)": next_row["start_time(ms)"],
                    "during_time(ms)": during_time,
                }
                new_rows.append(all_time_row)

            return pd.DataFrame(new_rows)

        all_time_dfs = batch_rows.groupby("pid").apply(create_all_time_rows).reset_index(drop=True)
        # 合并all_time行
        result_df = pd.concat([framework_df, all_time_dfs], ignore_index=True)
        result_df = result_df.sort_values(by=["start_time(ms)", "name"], ascending=[True, False]).reset_index(drop=True)

        return result_df
    
    def process_exporter(self, framework_df, name):
        # 划分组
        result_df = self._get_groups(framework_df, name)
        len_result_df = len(result_df)

        if len_result_df == 0:
            if self.args.batch_size > 0:
                size_recommend = self._get_batch_size_recommend(framework_df, name)
                logger.warning("%s: no %s with batch_size %d" % (name, self.batch_start_name, self.args.batch_size))
                if size_recommend[0] == -1:
                    logger.warning("no %s data, please check." % name)
                else:
                    logger.warning("%s: recommend batch_size from data %s" % (name, 
                                    ', '.join(map(str, size_recommend))))
            elif self.args.rid != "-1":
                logger.warning("%s: no %s with rid %r" % (name, self.batch_start_name, self.args.rid))
            return pd.DataFrame()

        merged = pd.concat(result_df, ignore_index=True)
        return merged
    
    def _get_groups(self, framework_df, name):
        result_df = []
        rid = self.args.rid
        batch_size = self.args.batch_size
        batch_num = self.args.batch_num

        def filter_by_rid(sub_group):
            filtered = sub_group[(sub_group["name"] == self.batch_start_name) &
                                 (sub_group["batch_type"] == name)]
            # 转换 rid_list 元素为字符串
            filtered.loc[:, "rid_list"] = filtered["rid_list"].apply(
                lambda x: [str(i) for i in x] if isinstance(x, list) else []
            )
            return filtered[filtered["rid_list"].apply(lambda x: rid in x)]

        def filter_by_batch_size(sub_group):
            return sub_group[(sub_group["name"] == self.batch_start_name) &
                             (sub_group["batch_type"] == name) &
                             (sub_group["batch_size"] == str(batch_size))]
        
        groups = framework_df.groupby((framework_df["name"] == self.batch_start_name).cumsum())
        result_number = 0
        for _, group in groups:
            if result_number == batch_num:
                break
            if rid != "-1":
                batch_group = filter_by_rid(group)
            elif batch_size > 0:
                batch_group = filter_by_batch_size(group)
            else:
                continue

            if batch_group.empty:
                continue

            if name == PREFILL_NAME and not self._is_valid_prefill(batch_group, framework_df):
                continue
            result = self._get_full_batch(group, framework_df)

            if not result.empty:
                result = self._get_cacl_df(result, framework_df, name)
                result_df.append(result) 
                result_number += 1
   
        return result_df

    def _is_valid_prefill(self, batch_group, framework_df):
        # batch_group不会为空且包含rid_list列
        batch_row = batch_group.iloc[0]
        cur_rid = batch_row["rid_list"][0]
        if self.args.rid != "-1":
            cur_rid = self.args.rid

        target_encode = framework_df[(framework_df["rid"] == str(cur_rid)) & 
                                     (framework_df["name"] == self.http_start_name)]
        return not target_encode.empty

    def _get_full_batch(self, group, framework_df):
        start_index = self.name_list.index(self.batch_start_name)
        end_index = self.name_list.index(self.batch_end_name)
        all_time_rows = group[group["name"] == self.all_time_name]
        if all_time_rows.empty:
            logger.debug(f"No row named {self.all_time_name} found in the group, skip this batch")
            return pd.DataFrame()
        all_time_index = all_time_rows.index[0]

        batch_start_rows = group[group["name"] == self.batch_start_name]
        if batch_start_rows.empty:
            logger.debug(f"No row named {self.batch_start_name} found in the group, skip this batch")
            return pd.DataFrame()
        batch_start_index = batch_start_rows.index[0]
        batch_rid = batch_start_rows.iloc[0]["rid"]
        
        concat_list = [batch_start_index, all_time_index]
        full_batch = self.name_list[start_index + 1: end_index + 1]

        # 找到key_row
        key_pid, key_tid = self._get_key_info(framework_df, batch_rid)
        if key_pid is None:
            logger.debug(f"no named {self.key_name} line, skip this batch")
            return pd.DataFrame()

        # 获取完整的batch
        result_index = self._get_batch_index(full_batch, batch_start_index, framework_df, key_pid, key_tid)
        if result_index.empty:
            return pd.DataFrame()
        framework_df.loc[framework_df["name"] == self.all_time_name, ["start_time(ms)", "end_time(ms)"]] = \
            framework_df.loc[framework_df["name"] == self.all_time_name, ["end_time(ms)", "start_time(ms)"]].values
        result = pd.concat([framework_df.loc[concat_list], result_index])
        result = result.sort_values(by=["start_time(ms)", "name"], ascending=[True, False]).reset_index(drop=True)
        return result
    
    def _get_key_info(self, framework_df, batch_rid):
        key_mask = ((framework_df["name"] == self.key_name) | (framework_df["name"] == "preprocess")) & \
                    (framework_df["rid"] == batch_rid)
        if not key_mask.any():
            logger.debug(f"no {self.key_name} line, skip this batch")
            return None, None
        key_row = framework_df[key_mask].iloc[0]
        key_pid, key_tid = key_row["pid"], key_row["tid"]
        return key_pid, key_tid
    
    def _get_cacl_df(self, filter_df, framework_df, name):
        empty_row = pd.DataFrame(index=[0])
        cur_rid = filter_df.iloc[0]["rid_list"][0]
        if self.args.rid != "-1":
            cur_rid = self.args.rid
        if name == PREFILL_NAME:
            http_df = framework_df[(framework_df["rid"] == str(cur_rid)) & 
                                (framework_df["name"].isin(self.http_list))]
            filter_df = pd.concat([filter_df, http_df], ignore_index=True)
            filter_df = filter_df.drop_duplicates(subset="name")
            filter_df = filter_df.sort_values(by="start_time(ms)")
        filter_df_ = filter_df[~filter_df["name"].isin(self.filter_list)]
        # 2. 当前行与下一行计算during_time
        add_df = self._calc_during_time(filter_df_)
        cur_df = pd.concat([filter_df, add_df], ignore_index=True)
        
        # 3. 与AllTime行的计算逻辑
        cur_df = self._postprocess_framework_df(cur_df, name)
        concat_df = pd.concat([empty_row, cur_df], ignore_index=True)
        return concat_df
    
    def _postprocess_framework_df(self, framework_df, name):
        if name == PREFILL_NAME:
            post_event = pd.concat([
                framework_df[framework_df["name"] == self.http_end_name],
                framework_df[framework_df["name"] == self.all_time_name]
            ])
        else:
            filter_df = framework_df[(framework_df["name"] != self.all_time_name) &
                                     (~framework_df["name"].str.startswith("Between-"))]
            if filter_df.empty:
                return pd.DataFrame()
            
            last_row = filter_df.iloc[[-1]]
            post_event = pd.concat([
                last_row,
                framework_df[framework_df["name"] == self.all_time_name]
            ])
        framework_df = framework_df.sort_values(by="start_time(ms)")
        new_rows = self._calc_during_time(post_event)

        framework_df = pd.concat([framework_df, new_rows], ignore_index=True)
        all_time_row = framework_df[framework_df["name"] == self.all_time_name]
        non_all_time_row = framework_df[framework_df["name"] != self.all_time_name]
        framework_df = pd.concat([non_all_time_row, all_time_row], ignore_index=True)
        
        return framework_df
   
    def _get_batch_size_recommend(self, framework_df, name):
        batch_df = framework_df[(framework_df["name"] == self.batch_start_name) &
                                (framework_df["batch_type"] == name)]
        if batch_df.empty:
            return [-1]
        batch_size = batch_df["batch_size"].unique()
        if len(batch_size) == 0:
            logger.warning(f"{name}: The batch_size is empty")
            return [-1]
        return batch_size


# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from abc import abstractmethod

from .base import BaseConstraint, ConstraintStatus
from ..utils import GlobalConfig


class LogicConstraint(BaseConstraint):
    @abstractmethod
    def _is_satisfied_by(self, val):
        pass

    def is_satisfied_by(self, val):
        if GlobalConfig.is_custom_set():
            return GlobalConfig.custom_return

        ret = self._is_satisfied_by(val)
        self.status = ConstraintStatus.SUCCESS if ret else ConstraintStatus.FAILURE
        return ret


class NullConstraint(LogicConstraint):
    def __init__(self, *, description=None):
        description = "always true" if description is None else description
        super().__init__(description=description)

    def _is_satisfied_by(self, val):
        return True


class FunctionConstraint(LogicConstraint):
    def __init__(self, func, description=None):
        super().__init__(description=description)

        # Check if the function has exactly one parameter
        sig = inspect.signature(func)
        if len(sig.parameters) != 1:
            raise ValueError(f"The function {func.__qualname__} must have exactly one parameter.")

        self.func = func
        self.description = description or func.__qualname__

    def _is_satisfied_by(self, val):
        result = self.func(val)

        if not isinstance(result, bool):
            raise TypeError(
                f"Expected 'result' to be bool. Got {type(result).__name__} instead."
            )
        return result


class AndConstraint(LogicConstraint):
    def __init__(self, *constraints, description=None):
        self.constraints = constraints
        description = " and ".join(c.description for c in self.constraints)
        super().__init__(description=description)

    def __str__(self):
        for constraint in self.constraints:
            if constraint.status == ConstraintStatus.FAILURE:
                return str(constraint)
        return self.description

    def _is_satisfied_by(self, val):
        return all(c.is_satisfied_by(val) for c in self.constraints)


class OrConstraint(LogicConstraint):
    def __init__(self, *constraints, description=None):
        self.constraints = constraints
        description = " or ".join(c.description for c in self.constraints)
        super().__init__(description=description)
        
    def __str__(self):
        for constraint in self.constraints:
            if constraint.status == ConstraintStatus.FAILURE:
                return str(constraint)
        return self.description

    def _is_satisfied_by(self, val):
        return any(c.is_satisfied_by(val) for c in self.constraints)
    
    
class NotConstraint(LogicConstraint):
    def __init__(self, constraint, *, description=None):
        self.constraint = constraint
        description = f"not {self.constraint.description}"
        super().__init__(description=description)

    def _is_satisfied_by(self, val):
        return not self.constraint.is_satisfied_by(val)


class IfElseConstraint(LogicConstraint):
    def __init__(self, condition, if_constraint, else_constraint, *, description=None):
        super().__init__(description=description)

        self.condition = make_constraint(condition, description)
        self.if_constraint = if_constraint
        self.else_constraint = else_constraint
        self.condition_res = False
    
    def __str__(self):
        return f"{self.if_constraint if self.condition_res else self.else_constraint}"

    def _is_satisfied_by(self, val):
        self.condition_res = self.condition.is_satisfied_by(val)
        return self.if_constraint.is_satisfied_by(val) if \
               self.condition_res else \
               self.else_constraint.is_satisfied_by(val)


def make_constraint(constraint, description=None) -> BaseConstraint:
    """Convert None, bool, function and constraint to constraint"""
    if constraint is None:
        return NullConstraint()

    if isinstance(constraint, BaseConstraint):
        return constraint
    
    if isinstance(constraint, bool):
        if description is None:
            raise ValueError("'description' must not be None when 'condition' is bool")

        return FunctionConstraint(lambda _: constraint, description)

    if callable(constraint):
        return FunctionConstraint(constraint, description)

    raise TypeError(
        f"Expected 'constraint' to be BaseContraint, bool, or Callable. "
        f"Got {type(constraint).__name__} instead."
    )


def where(condition, if_constraint, else_constraint, *, description=None):
    """description is used for 'condition'"""
    return IfElseConstraint(
        condition,
        if_constraint,
        else_constraint,
        description=description
    )



# -*- coding: utf-8 -*-
# Copyright (c) Huawei Technologies Co., Ltd. 2023-2024. All rights reserved.

import copy
import json
import statistics
from pathlib import Path
from statistics import mean, stdev
from dataclasses import dataclass
from typing import Dict, List, Optional

import numpy as np
from matplotlib import pyplot as plt
from loguru import logger
from msserviceprofiler.msguard.security import open_s
from msserviceprofiler.modelevalstate.common import State


@dataclass
class PlotConfig:
    data: Dict[State, List]
    x_field: str
    title: str
    x_label: str
    y_label: str
    save_path: str = None


class AnalysisState:

    @staticmethod
    def computer_mean_sigma(data: Dict[State, List], x_field: str, ):
        # 合并只有decode， prefill
        res = {}
        tmp_data = copy.deepcopy(data)
        for k, v in tmp_data.items():
            if k.batch_prefill:
                s = State(batch_prefill=k.batch_prefill)
            else:
                s = State(batch_decode=k.batch_decode)
            if s in res:
                res[s].extend(v)
            else:
                res[s] = v
        # 计算mean sigma
        _x = []
        _mean = []
        _positive_sigma = []
        _negative_sigma = []
        for k in sorted(res.keys(), key=lambda x: getattr(x, x_field)):
            v = res.get(k, [])
            if len(v) == 1:
                _x.append(getattr(k, x_field))
                _mean.append(v[0])
                _positive_sigma.append(v[0])
                _negative_sigma.append(v[0])
                continue
            _x.append(getattr(k, x_field))
            _mean.append(mean(v))
            try:
                _sigma = stdev(v)
            except AssertionError:
                try:
                    _sigma = np.std(v)
                except Exception:
                    logger.warning('Failed stdev', v)
                    _sigma = 0 if not v else statistics.pstdev(v)
            _positive_sigma.append(_mean[-1] + _sigma)
            _negative_sigma.append(_mean[-1] - _sigma)
        return _x, _mean, _positive_sigma, _negative_sigma

    @staticmethod
    def plot_input_velocity(config: PlotConfig):
        """
        绘制输入数据的平均值，上波动 和下波动曲线。
        :return:
        """
        # 合并只有decode， prefill
        _x, _mean, _positive_sigma, _negative_sigma = AnalysisState.computer_mean_sigma(config.data, config.x_field)
        plt.plot(_x, _mean, label="mean")
        plt.plot(_x, _positive_sigma, label="positive std")
        plt.plot(_x, _negative_sigma, label="negative std")
        plt.title(config.title)
        plt.legend()
        plt.grid()
        if config.x_label:
            plt.xlabel(config.x_label)
        if config.y_label:
            plt.ylabel(config.y_label)
        if config.save_path:
            plt.savefig(Path(config.save_path).joinpath(f"{config.x_label}_{config.y_label}_{config.title}.png"))
            plt.close()
        else:
            plt.show()

    @staticmethod
    def plot_input_velocity_with_df(predict_df, origin_df, save_path=None):
        """
        绘制输入数据的平均值，上波动 和下波动曲线。
        :return:
        """
        # 合并只有decode， prefill
        _batch_stage = predict_df["batch_stage"].unique()
        for _bs in _batch_stage:
            _cur_predict_df = predict_df[predict_df["batch_stage"] == _bs]
            _cur_origin_df = origin_df[origin_df["batch_stage"] == _bs]
            _p_group = _cur_predict_df.groupby("batch_size")
            _p_mean = _p_group.mean()
            _p_sigma = _p_group.std().fillna(0)
            _o_mean = _cur_origin_df.groupby("batch_size").mean().values.flatten()
            _o_sigma = _cur_origin_df.groupby("batch_size").std().fillna(0).values.flatten()

            plt.plot(_p_mean.index.values, _p_mean.values.flatten(), label="predict mean")
            plt.plot(_p_mean.index.values, (_p_mean.values + _p_sigma.values).flatten(), label="predict positive std")
            plt.plot(_p_mean.index.values, (_p_mean.values - _p_sigma.values).flatten(), label="predict negative std")

            plt.plot(_o_mean.index.values, _o_mean.values.flatten(), label="origin mean")
            plt.plot(_o_mean.index.values, (_o_mean.values + _o_sigma.values).flatten(), label="origin positive std")
            plt.plot(_o_mean.index.values, (_o_mean.values - _o_sigma.values).flatten(), label="origin negative std")
            plt.title(f"{_bs} latency")
            plt.legend()
            plt.grid()
            plt.xlabel("batch size")
            plt.ylabel("res")
            if save_path:
                plt.savefig(Path(save_path).joinpath(f"{_bs}_batch_size_res.png"))
                plt.close()
            else:
                plt.show()

    @staticmethod
    def plot_input_velocity_with_predict(config: PlotConfig, predict_data: Dict[State, List]):
        """
        绘制输入数据和预测数据的平均值，上波动 和下波动曲线。
        :return:
        """
        # 合并只有decode， prefill
        # 合并只有decode， prefill
        _x, _mean, _positive_sigma, _negative_sigma, = AnalysisState.computer_mean_sigma(config.data, config.x_field)
        plt.figure()
        plt.plot(_x, _mean, label="mean")
        plt.plot(_x, _positive_sigma, label="positive std")
        plt.plot(_x, _negative_sigma, label="negative std")
        _x, _predict, _predict_positive_sigma, _predict_negative_sigma = AnalysisState.computer_mean_sigma(
            predict_data,
            config.x_field
        )

        plt.plot(_x, _predict, label="predict")
        plt.plot(_x, _predict_positive_sigma, label="predict positive std")
        plt.plot(_x, _predict_negative_sigma, label="predict negative std")
        plt.title(config.title)
        plt.legend()
        plt.grid()
        if config.x_label:
            plt.xlabel(config.x_label)
        if config.y_label:
            plt.ylabel(config.y_label)
        if config.save_path:
            plt.savefig(Path(config.save_path).joinpath(f"{config.x_label}_{config.y_label}_{config.title}.png"))
            plt.close()
            with open_s(config.save_path.joinpath(f"{config.title}.txt"), 'w') as f:
                f.write('mean\n')
                f.write(json.dumps(_mean))
                f.write('\n')
                f.write('positive std\n')
                f.write(json.dumps(_positive_sigma))
                f.write('\n')
                f.write('negative std\n')
                f.write(json.dumps(_negative_sigma))
                f.write('\n')
                f.write('predict \n')
                f.write(json.dumps([float(i) for i in _predict]))
        else:
            plt.show()
        

    @staticmethod
    def plot_pred_and_real(pred, real, save_path: Optional[Path] = None):
        plt.figure()
        plt.scatter(range(len(pred)), pred, label='pred', alpha=0.5)
        plt.scatter(range(len(real)), real, label="real", alpha=0.5)
        plt.title("predict value and real value")
        plt.xlabel("index")
        plt.ylabel("value")
        plt.legend()
        if save_path:
            plt.savefig(save_path.joinpath("predict value and real value.png"))
            plt.close()
        else:
            plt.show()



# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from itertools import cycle
from math import isinf, isnan
from typing import Optional, Tuple

import numpy as np

from msserviceprofiler.modelevalstate.config.config import default_support_field, PerformanceIndex, \
    map_param_with_value, OptimizerConfigField
from msserviceprofiler.modelevalstate.config.base_config import REQUESTRATES, CONCURRENCYS


class StopFineTune(Exception):
    pass


class FineTune:
    def __init__(self, ttft_penalty: float = 0, tpot_penalty: float = 0, target_field: Optional[Tuple] = None,
                 ttft_slo: float = 0.5, tpot_slo: float = 0.05, slo_coefficient: float = 0.1, step_size: float = 0.5):
        self.ttft_penalty = ttft_penalty  # 优化算法中惩罚系数
        self.tpot_penalty = tpot_penalty
        self.ttft_slo = ttft_slo
        self.tpot_slo = tpot_slo
        self.slo_coefficient = slo_coefficient
        self.target_field = target_field if target_field else default_support_field
        self.fine_tune_target = ["REQUESTRATE"]
        self.fine_tune_type = cycle(self.fine_tune_target)
        self.step_size = step_size
        self.ttft_lower_bound = self.ttft_slo * (1 - self.slo_coefficient)
        self.ttft_upper_bound = self.ttft_slo
        self.tpot_lower_bound = self.tpot_slo * (1 - self.slo_coefficient)
        self.tpot_upper_bound = self.tpot_slo
        if self.ttft_penalty == 0 and self.tpot_penalty == 0:
            raise StopFineTune("No penalties, no need to fine-tune.")
        ttft_flag = self.ttft_penalty != 0 and self.ttft_slo == 0
        tpot_flag = self.tpot_penalty != 0 and self.tpot_slo == 0
        if ttft_flag or tpot_flag:
            raise ValueError("Penalty is set but SLO is zero.")
        self.ttft_over_slo = False
        self.ttft_under_lower_bound = False
        self.tpot_over_slo = False
        self.tpot_under_lower_bound = False
        self.last_signed_factor = {}
        self.last_value = {}

    @staticmethod
    def update_field(simulate_run_info, signed_factor, field_names: tuple = REQUESTRATES, 
                     last: Optional[float] = None) -> bool:
        if signed_factor == 0 or isinf(signed_factor) or isnan(signed_factor):
            return False
        for _field in simulate_run_info:
            if _field.name.upper().strip() in field_names:
                if _field.constant is not None or _field.min == _field.max:
                    return False
                original_value = _field.value
                if last:
                    _new_value = _field.value + signed_factor * abs(_field.value - last)
                else:
                    _new_value = _field.value * (1 + signed_factor)
                if isinf(_new_value) or isnan(_new_value):
                    return False
                _field.value = _new_value
                _new_value = max(_field.min, min(_field.max, _field.value))
                if isinf(_new_value) or isnan(_new_value):
                    _field.value = original_value
                    return False
                _field.value = _new_value
                # 检查值是否发生了有影响的变化(>=0.1)
                return abs(_field.value - original_value) >= 0.1
        return False
    
    @staticmethod
    def add_history(target, key_name, value):
        if key_name in target:
            target[key_name].append(value)
        else:
            target[key_name] = [value]

    def reset_history(self):
        self.last_signed_factor = {}
        self.last_value = {}
    
    def check_config_and_performance(self, performance_index: PerformanceIndex):

        if self.ttft_penalty == 0 and self.tpot_penalty == 0:
            raise StopFineTune("No penalties, no need to fine-tune.")
        ttft_flag = self.ttft_penalty != 0 and self.ttft_slo == 0
        tpot_flag = self.tpot_penalty != 0 and self.tpot_slo == 0
        if ttft_flag or tpot_flag:
            raise ValueError("Penalty is set but SLO is zero.")
        if performance_index.time_per_output_token is None:
            raise ValueError("Missing performance data for TPOT.")
        if self.ttft_penalty != 0 and performance_index.time_to_first_token is None:
            raise ValueError("Missing performance data for TTFT.")
        
    def direction_of_field_update(self, performance_index: PerformanceIndex):
        actual_tpot = performance_index.time_per_output_token
        actual_ttft = performance_index.time_to_first_token
        self.ttft_over_slo = False
        self.ttft_under_lower_bound = False
        self.tpot_over_slo = actual_tpot > self.tpot_upper_bound
        self.tpot_under_lower_bound = actual_tpot < self.tpot_lower_bound
        # 同时约束ttft
        if self.ttft_penalty != 0:
            self.ttft_over_slo = actual_ttft > self.ttft_upper_bound
            self.ttft_under_lower_bound = actual_ttft < self.ttft_lower_bound

    def handle_concurrency(self, simulate_run_info: Tuple[OptimizerConfigField, ...],
                           performance_index: PerformanceIndex):
        was_updated_c = False
        signed_factor_c = None
        if self.tpot_over_slo:
            deviation_ratio = (performance_index.time_per_output_token - self.tpot_slo) / self.tpot_slo
            signed_factor_c = -deviation_ratio * self.step_size
        elif self.tpot_under_lower_bound:
            deviation_ratio = (self.tpot_slo - performance_index.time_per_output_token) / self.tpot_slo
            signed_factor_c = deviation_ratio * self.step_size
        # 不满足 调整并发
        if signed_factor_c:
            self.add_history(self.last_signed_factor, CONCURRENCYS, signed_factor_c)
            signed_factor_c = max(-self.step_size, min(self.step_size, signed_factor_c))
            _concurrency_signed_factor = self.last_signed_factor.get(CONCURRENCYS)
            last_concurrency = None
            if len(_concurrency_signed_factor or []) >= 2:
                # 表示上一次指标和本次指标的方向不一致，可往中间调节
                if (_concurrency_signed_factor[-2] * _concurrency_signed_factor[-1] < 0 and
                        self.last_value.get(CONCURRENCYS)[-2] != self.last_value.get(CONCURRENCYS)[-1]):
                    last_concurrency = self.last_value.get(CONCURRENCYS)[-2]
            was_updated_c = self.update_field(simulate_run_info, signed_factor_c, field_names=CONCURRENCYS,
                                              last=last_concurrency)
        return was_updated_c
 
    def handle_request_rate(self, simulate_run_info: Tuple[OptimizerConfigField, ...],
                            performance_index: PerformanceIndex):
        # 检查ttft是否满足，调整request rate
        was_updated_r = False
        signed_factor_r = None
        if self.ttft_over_slo:
            deviation_ratio = (performance_index.time_to_first_token - self.ttft_slo) / self.ttft_slo
            signed_factor_r = -deviation_ratio * self.step_size
            signed_factor_r = max(-self.step_size, min(self.step_size, signed_factor_r))
        elif self.ttft_under_lower_bound:
            signed_factor_r = self.step_size
        if signed_factor_r:
            self.add_history(self.last_signed_factor, REQUESTRATES, signed_factor_r)
            # 限制最大更新幅度
            _request_rate_factor = self.last_signed_factor.get(REQUESTRATES)
            last_req_rate = None
            if len(_request_rate_factor or []) >= 2:
                # 表示上一次指标和本次指标的方向不一致，可往中间调节
                if _request_rate_factor[-2] * _request_rate_factor[-1] < 0:
                    last_req_rate = self.last_value.get(REQUESTRATES)[-2]
            was_updated_r = self.update_field(simulate_run_info, signed_factor_r, field_names=REQUESTRATES,
                                              last=last_req_rate)
        return was_updated_r
 
    def fine_tune_with_concurrency_and_request_rate(self, params: np.ndarray, performance_index: PerformanceIndex):
        # 第一次 获取到的是并发， request rate 最小
        self.check_config_and_performance(performance_index)
        self.direction_of_field_update(performance_index)
        simulate_run_info = map_param_with_value(params, self.target_field)
        _concurrency_flag = _request_rate_flag = False
        _concurrency_field = _request_field = None
        for _field in simulate_run_info:
            if _field.name in REQUESTRATES:
                self.add_history(self.last_value, REQUESTRATES, _field.value)
                _request_rate_flag = True
                _request_field = _field
            if _field.name in CONCURRENCYS:
                self.add_history(self.last_value, CONCURRENCYS, _field.value)
                _concurrency_flag = True
                _concurrency_field = _field
        if _concurrency_flag:
            was_updated_c = self.handle_concurrency(simulate_run_info, performance_index)
            if was_updated_c:
                # 当并发 更新了，request rate 恢复为最初的request rate的value。
                if _concurrency_field.value != self.last_value[CONCURRENCYS][-1] and _request_field:
                    _request_field.value = self.last_value[REQUESTRATES][0]
                return simulate_run_info
        if _request_rate_flag:
            # 更新request rate 上下限
            for _field in simulate_run_info:
                if _field.name in REQUESTRATES:
                    if _concurrency_flag and self.last_value.get(CONCURRENCYS)[-1] < _field.max:
                        _field.max = self.last_value.get(CONCURRENCYS)[-1]
            # 检查ttft是否满足，调整request rate
            was_updated_r = self.handle_request_rate(simulate_run_info, performance_index)
            if was_updated_r:
                return simulate_run_info
        raise StopFineTune("Parameter value reached its boundary or did not change.")


# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import traceback
import importlib
import inspect
import functools
import sys
from abc import ABC, abstractmethod
from typing import Union, Tuple, List, Optional, Callable, Any, Set
from packaging.version import Version
from .utils import logger
from .registry import get_hook_registry, add_to_hook_registry


def import_object_from_string(import_path: str, module_path: str) -> Any:
    """
    根据点分隔的路径导入对象（模块、类、函数等）
    支持多级嵌套属性，如 "module.submodule.ClassName.method_name"
    """
    if not import_path:
        logger.error(f"Module import_path is empty")
        return None

    try:
        module = importlib.import_module(import_path)
    except ImportError as e:
        logger.error(f"Module import failed for %s: %s", module_path, e)
        return None

    current = module
    for part in module_path.split("."):
        prev = current  # For log
        current = getattr(current, part, None)
        if current is None:
            logger.error(f"Module {prev} doesn't has attribute {part}")
            return None
    return current


class HookHelper:
    """辅助类，用于函数替换操作"""
    def __init__(self, ori_function_define, new_function):
        self.new_function = new_function
        self.ori_function = None
        self.location = None
        self.attr_name = None

        if ori_function_define is None:
            return

        if inspect.isfunction(ori_function_define) or inspect.ismethod(ori_function_define):
            self.ori_function = ori_function_define
            self.location, self.attr_name = self.get_location(self.ori_function)
        elif callable(ori_function_define):
            self.ori_function = ori_function_define.__call__
            self.location, self.attr_name = self.get_location(self.ori_function)

        # Allow initialization without new_function; validation will occur in replace()/recover()
        if not all([self.ori_function, self.location, self.attr_name]):
            warn_msg = f"{ori_function_define} initialization failed."
            logger.error(warn_msg)
            raise ValueError(warn_msg)

    @staticmethod
    def get_location(function_ins):
        if not hasattr(function_ins, "__module__"):
            warning_msg = f"function {str(function_ins)} has no __module__."
            logger.error(warning_msg)
            raise ValueError(warning_msg)

        module = importlib.import_module(function_ins.__module__)
        qualified_name = function_ins.__qualname__.split(".")
        classes = qualified_name[:-1]
        attr_name = qualified_name[-1]
        location = module

        for class_name in classes:
            location = getattr(location, class_name, None)
            if location is None:
                break

        if location is None:
            warning_msg = f"{'.'.join(classes)} does not exist"
            logger.error(warning_msg)
            raise ValueError(warning_msg)

        return location, attr_name

    def replace(self):
        if all([self.ori_function, self.location, self.attr_name, self.new_function]):
            try:
                setattr(self.location, self.attr_name, self._get_method(self.new_function))
                logger.debug(f"Successfully replaced {self.attr_name} in {self.location}")
            except Exception as e:
                logger.error(f"Failed to replace {self.attr_name} in {self.location}: {e}")
                raise

    def recover(self):
        if all([self.ori_function, self.location, self.attr_name, self.new_function]):
            try:
                setattr(self.location, self.attr_name, self._get_method(self.ori_function))
                logger.debug(f"Successfully recovered {self.attr_name} in {self.location}")
            except Exception as e:
                logger.error(f"Failed to recover {self.attr_name} in {self.location}: {e}")
                raise

    def _get_method(self, func):
        if inspect.isclass(self.location):
            try:
                func_cls_name = inspect.getattr_static(self.location, self.attr_name).__class__.__name__
                if func_cls_name in ("staticmethod", "classmethod"):
                    return staticmethod(func)
            except AttributeError:
                pass
        return func


def get_parents_name(ori_func, index=1):
    """获取调用栈中的父级函数名"""
    gen = traceback.walk_stack(None)
    try:
        for _ in range(index + 1):
            f = next(gen)
        return f[0].f_code.co_name
    except StopIteration:
        return None


class VLLMHookerBase(ABC):
    """Hooker基类，提供核心功能"""
    vllm_version = (None, None)  # (min_version, max_version)
    applied_hook_func_name = ""

    def __init__(self):
        self.hooks = []
        # 记录文本形式的 hook 点（未解析成对象前的 (import_path, attr_path) 列表）
        self.hook_list: List[Tuple[str, str]] = []
        self.caller_filter: Optional[str] = None
        # 对应的 hook 处理函数，用于配置化时复用
        self.hook_func: Optional[Callable] = None

    @abstractmethod
    def init(self):
        """初始化hook点，子类必须实现"""
        pass

    def replace_func(self, ori_func, pname, profiler_func, on_recover: Optional[Callable] = None):
        """创建替换函数，带失败回退机制。

        约定：
        - 任一 hook 执行抛异常时，不影响业务逻辑：返回原始函数结果；
        - 连续失败达到 5 次，自动回退为原始函数（若提供 on_recover 则调用以执行恢复）。
        - 支持 context manager 的特殊处理。
        """
        failures = 0
        # 检测是否为 context manager 装饰的函数
        # 检查 profiler_func 是否是生成器函数（@contextmanager 装饰的函数）
        is_context_manager_func = inspect.isgeneratorfunction(profiler_func)
        # 检测是否为异步函数
        is_async_func = inspect.iscoroutinefunction(ori_func) or inspect.iscoroutinefunction(profiler_func)

        if is_async_func:
            @functools.wraps(ori_func)
            async def async_wrapper(*args, **kwargs):
                if pname is not None and get_parents_name(ori_func) != pname:
                    logger.debug(f"calling {ori_func}")
                    return await ori_func(*args, **kwargs)

                nonlocal failures
                if failures >= 5:
                    return await ori_func(*args, **kwargs)

                try:
                    logger.debug(f"calling profiler_func={self.applied_hook_func_name} for {ori_func}")
                    return await profiler_func(*args, **kwargs)
                except Exception as e:
                    failures += 1
                    logger.error(f"Hook execution failed ({failures}/5) for {ori_func}: {e}")
                    logger.error(f"Exception details: {type(e).__name__}: {str(e)}")
                    import traceback
                    logger.error(f"Traceback: {traceback.format_exc()}")
                    
                    # 对于 context manager 函数，第一次失败就立即回退
                    if is_context_manager_func:
                        logger.warning(f"Context manager hook failed, immediately reverting to original for {ori_func}")
                        try:
                            if callable(on_recover):
                                on_recover()
                        except Exception as re:
                            logger.error(f"Failed to recover original function for {ori_func}: {re}")
                        # 对于 context manager，直接返回原始函数的结果
                        return await ori_func(*args, **kwargs)
                    
                    # 异步函数，调用原始函数获取结果
                    try:
                        result = await ori_func(*args, **kwargs)
                    except Exception as orig_e:
                        logger.error(f"Original function also failed for {ori_func}: {orig_e}")
                        # 如果原始函数也失败，重新抛出异常
                        raise orig_e
                    
                    # 达到5次失败阈值时回退
                    if failures >= 5:
                        try:
                            if callable(on_recover):
                                on_recover()
                            logger.warning(f"Hook permanently reverted to original for {ori_func} after 5 failures")
                        except Exception as re:
                            logger.error(f"Failed to recover original function for {ori_func}: {re}")
                    return result
            return async_wrapper
        else:
            @functools.wraps(ori_func)
            def wrapper(*args, **kwargs):
                if pname is not None and get_parents_name(ori_func) != pname:
                    logger.debug(f"calling {ori_func}")
                    return ori_func(*args, **kwargs)

                nonlocal failures
                if failures >= 5:
                    return ori_func(*args, **kwargs)

                try:
                    logger.debug(f"calling profiler_func={self.applied_hook_func_name} for {ori_func}")
                    return profiler_func(*args, **kwargs)
                except Exception as e:
                    failures += 1
                    logger.error(f"Hook execution failed ({failures}/5) for {ori_func}: {e}")
                    logger.error(f"Exception details: {type(e).__name__}: {str(e)}")
                    import traceback
                    logger.error(f"Traceback: {traceback.format_exc()}")
                    
                    # 对于 context manager 函数，第一次失败就立即回退
                    if is_context_manager_func:
                        logger.warning(f"Context manager hook failed, immediately reverting to original for {ori_func}")
                        try:
                            if callable(on_recover):
                                on_recover()
                        except Exception as re:
                            logger.error(f"Failed to recover original function for {ori_func}: {re}")
                        # 对于 context manager，直接返回原始函数的结果
                        return ori_func(*args, **kwargs)
                    
                    # 普通函数，调用原始函数获取结果
                    try:
                        result = ori_func(*args, **kwargs)
                    except Exception as orig_e:
                        logger.error(f"Original function also failed for {ori_func}: {orig_e}")
                        # 如果原始函数也失败，重新抛出异常
                        raise orig_e
                    
                    # 达到5次失败阈值时回退
                    if failures >= 5:
                        try:
                            if callable(on_recover):
                                on_recover()
                            logger.warning(f"Hook permanently reverted to original for {ori_func} after 5 failures")
                        except Exception as re:
                            logger.error(f"Failed to recover original function for {ori_func}: {re}")
                    return result
            return wrapper

    def do_hook(self, hook_points, profiler_func_maker, pname=None):
        """执行实际的hook操作"""
        for ori_func in hook_points:
            if ori_func is None:
                continue
            profiler_func = profiler_func_maker(ori_func)
            cur_hook = HookHelper(ori_func, None)
            def _recover_current(cur_hook_ref=cur_hook):
                try:
                    cur_hook_ref.recover()
                except Exception as e:
                    logger.error(f"Recover call failed: {e}")
            wrapped = self.replace_func(ori_func, pname, profiler_func, on_recover=_recover_current)
            cur_hook.new_function = wrapped
            cur_hook.replace()
            self.hooks.append(cur_hook)
            logger.info(f"Successfully applied hook wrapper for {ori_func} with {self.applied_hook_func_name}")
            logger.debug(f"replacing {ori_func} with {self.applied_hook_func_name}")

    def support_version(self, version):
        """检查当前版本是否支持"""
        min_version = self.vllm_version[0]
        max_version = self.vllm_version[1]

        if min_version is not None and Version(min_version) > Version(version):
            logger.debug(f"min_version={min_version} is less than version={version}, skip")
            return False
        if max_version is not None and Version(max_version) < Version(version):
            logger.debug(f"max_version={max_version} is larger than version={version}, skip")
            return False
        return True

    def register(self):
        """注册hooker到全局注册表"""
        add_to_hook_registry(self)


def vllm_hook(
    hook_points: Union[Tuple[str, str], List[Tuple[str, str]]],
    min_version: Optional[str] = None,
    max_version: Optional[str] = None,
    caller_filter: Optional[str] = None
) -> Callable:
    """
    装饰器工厂函数，用于简化hooker创建
    :param hook_points: hook点列表，格式为(模块名, 属性路径)
    :param min_version: 支持的最小版本
    :param max_version: 支持的最大版本
    :param caller_filter: 调用者过滤条件
    """
    def decorator(hook_func):
        logger.debug(f"Handling {hook_func}")
        
        class AutoHooker(VLLMHookerBase):
            vllm_version = (min_version, max_version)
            applied_hook_func_name = getattr(hook_func, "__name__", str(hook_func))

            def init(self):
                hook_list = [hook_points] if isinstance(hook_points, tuple) else hook_points
                points = [import_object_from_string(import_path, func_path) for import_path, func_path in hook_list]
                self.do_hook(
                    hook_points=points,
                    profiler_func_maker=lambda ori_func: lambda *args, **kwargs: hook_func(ori_func, *args, **kwargs),
                    pname=caller_filter,
                )
        hooker = AutoHooker()
        # 在注册阶段即缓存文本点位与处理函数，便于配置化复用
        try:
            hook_list_local = [hook_points] if isinstance(hook_points, tuple) else list(hook_points)
        except Exception:
            hook_list_local = []
        hooker.hook_list = hook_list_local
        hooker.caller_filter = caller_filter
        hooker.hook_func = hook_func
        hooker.register()
        return hook_func

    return decorator


def apply_hooks(version: str = None, filter_points: Optional[Set[Tuple[str, str]]] = None):
    """应用注册的hookers
    :param version: vllm 版本，缺省时自动探测
    :param filter_points: 若提供，则仅应用其文本点位列表中出现的 hookers；
                          当某 hooker 含多个点位时，会整体应用（具体点位拆分请使用动态注册能力）。
    """
    if version is None:
        import vllm  # noqa: F401
        version = vllm.__version__
    for hooker in get_hook_registry():
        if not hooker.support_version(version):
            logger.debug(f"Skipping hooker: {hooker.__class__.__name__} for version not matched")
            continue
        if filter_points is not None:
            # 仅当 hooker 的任一文本点位命中筛选集合时才应用
            try:
                hook_list = getattr(hooker, "hook_list", []) or []
                if not any(point in filter_points for point in hook_list):
                    continue
            except Exception:
                # 兼容旧对象，若取不到 hook_list，则跳过筛选逻辑
                pass
        try:
            logger.info(f"Applying hooker: {hooker.__class__.__name__}")
            hooker.init()
            logger.info(f"Successfully applied hooker: {hooker.__class__.__name__}")
        except Exception as e:
            logger.error(f"Failed to apply hooker {hooker.__class__.__name__}: {str(e)}")
            import traceback
            logger.error(f"Traceback: {traceback.format_exc()}")
            # 继续处理其他 hooker，不中断整个流程
            continue


def install_global_exception_handler():
    """安装全局异常处理器，捕获未被处理的异常"""
    def handle_exception(exc_type, exc_value, exc_traceback):
        if issubclass(exc_type, KeyboardInterrupt):
            # 允许键盘中断正常传播
            sys.__excepthook__(exc_type, exc_value, exc_traceback)
            return
        
        logger.error(f"Uncaught exception: {exc_type.__name__}: {exc_value}")
        logger.error(f"Traceback: {traceback.format_exception(exc_type, exc_value, exc_traceback)}")
        
        # 调用默认的异常处理器
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
    
    sys.excepthook = handle_exception
    logger.info("Global exception handler installed")


# 自动安装全局异常处理器
install_global_exception_handler()



# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# 全局注册表存储所有hookers
HOOK_REGISTRY = []
# 仅用于在基于配置的场景下，临时保存已导入的内置 Hooker 列表
BUILTIN_HOOKERS_CACHE = []


def clear_hook_registry():
    global HOOK_REGISTRY
    HOOK_REGISTRY = []


def clear_builtin_cache():
    global BUILTIN_HOOKERS_CACHE
    BUILTIN_HOOKERS_CACHE = []


def get_hook_registry():
    return HOOK_REGISTRY


def get_builtin_cache():
    return BUILTIN_HOOKERS_CACHE


def add_to_hook_registry(hooker):
    HOOK_REGISTRY.append(hooker)


def set_builtin_cache(cache):
    global BUILTIN_HOOKERS_CACHE
    BUILTIN_HOOKERS_CACHE = cache



# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod


class BaseComparator(ABC):
    SUPPORTED_EXTENSIONS = []
    
    def __init__(self, out_db_conn, excel_writer):
        self.out_db_conn = out_db_conn
        self.excel_writer = excel_writer
    
    @classmethod
    def supports(cls, file_extension):
        return file_extension in cls.SUPPORTED_EXTENSIONS
    
    @abstractmethod
    def process(self, file_a, file_b):
        pass



# !/usr/bin/python3.8
# -*- coding: utf-8 -*-
# Copyright (c) Huawei Technologies Co., Ltd. 2023-2024. All rights reserved.
from typing import List, Optional

import numpy as np
from loguru import logger
from pyswarms.backend.operators import compute_pbest
from pyswarms.single.global_best import GlobalBestPSO


class CustomGlobalBestPSO(GlobalBestPSO):
    def __init__(self, *args, breakpoint_cost: Optional[List] = None, breakpoint_pos: Optional[List] = None, **kwargs):
        super(CustomGlobalBestPSO, self).__init__(*args, **kwargs)
        self.breakpoint_cost = breakpoint_cost
        self.breakpoint_pos = breakpoint_pos
        if self.breakpoint_pos and self.breakpoint_cost:
            self.computer_next_pos()

    def computer_next_pos(self):
        self.swarm.pbest_cost = np.full(self.swarm_size[0], np.inf)
        if self.n_particles == 0:
            raise ValueError("n_particles cannot be zero")
        _iter_num = len(self.breakpoint_pos) // self.n_particles
        if (len(self.breakpoint_pos) % self.n_particles) != 0:
            _iter_num += 1
        for i in range(_iter_num):
            _current_pos = np.array(self.breakpoint_pos[i * self.n_particles:(i + 1) * self.n_particles])
            if _current_pos.shape[0] < self.n_particles:
                _current_pos = np.append(_current_pos, self.swarm.position[_current_pos.shape[0]:], axis=0)
            _current_cost = np.array(self.breakpoint_cost[i * self.n_particles:(i + 1) * self.n_particles])
            if _current_cost.shape[0] < self.n_particles:
                if self.swarm.current_cost.shape[0] != 0:
                    _current_cost = np.append(_current_cost, self.swarm.current_cost[_current_cost.shape[0]:], axis=0)
                else:
                    _current_cost = np.append(_current_cost, self.swarm.pbest_cost[_current_cost.shape[0]:], axis=0)
            self.swarm.position = _current_pos
            self.swarm.current_cost = _current_cost
            self.swarm.pbest_pos, self.swarm.pbest_cost = compute_pbest(self.swarm)
            self.swarm.best_pos, self.swarm.best_cost = self.top.compute_gbest(self.swarm)

            vel = self.swarm.velocity
            pos = self.swarm.position
            cost = self.swarm.best_cost
            mean_neighbor = self.swarm.best_cost
            pcost = np.mean(self.swarm.pbest_cost)
            hist_ = self.ToHistory(velocity=vel, position=pos, best_cost=cost, mean_pbest_cost=pcost,
                mean_neighbor_cost=self.swarm.best_cost,)
            self._populate_history(hist_)

        # Perform velocity and position updates
        self.swarm.velocity = self.top.compute_velocity(self.swarm, self.velocity_clamp, self.vh, self.bounds)
        dtype = self.swarm.velocity.dtype
        self.swarm.position = self.swarm.position.astype(dtype)
        self.swarm.position = self.top.compute_position(self.swarm, self.bounds, self.bh)
        logger.info(f"Best Position {self.swarm.best_pos}, Best Cost {self.swarm.best_cost}")
        logger.info(f"Init Position {self.swarm.position}")


# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class InvalidParameterError(Exception):
    # Class-level constants for colors (only computed once)
    _COLORS = {
        'RED': "\033[1;31m",
        'YELLOW': "\033[1;33m",
        'CYAN': "\033[1;36m",
        'RESET': "\033[0m"
    }

    def __init__(self, parameter_name, caller_name, constraint, parameter_value):
        self.parameter_name = parameter_name
        self.caller_name = caller_name
        self.constraint = constraint
        self.parameter_value = parameter_value

        super().__init__(self.build_error_message())

    def build_error_message(self):
        """Build the complete error message with colors and formatting"""
        message = f"Parameter validation failed in " \
                  f"{self._COLORS['CYAN']}function {self.caller_name!r}{self._COLORS['RESET']}: " \
                  f"{self._COLORS['YELLOW']}expected parameter {self.parameter_name!r} " \
                  f"{self.constraint}{self._COLORS['RESET']}, " \
                  f"{self._COLORS['RED']}but receieved {self.parameter_value!r}{self._COLORS['RESET']}"

        return message



# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

PLUGIN_WHITE_LIST = [*PLUGIN_WHITE_LIST, 'simulate']


class SimulatePluginParameterValidator(PluginParameterValidator):
    def __init__(self, speculation_gamma):
        super().__init__(speculation_gamma)
        self.rules['simulate'] = {
            PLUGIN_FIELDS: set(),  # 没有额外的字段要求
            PLUGIN_CHECK_FUNC: lambda data: True  # 不需要额外校验
        }


PluginParameterValidator = SimulatePluginParameterValidator
